import vak.config


def test_config_attrs_class(
    all_generated_configs_toml_path_pairs,
    default_model,
):
    """test that instantiating Config class works as expected"""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if default_model not in str(toml_path):
            continue  # only need to check configs for one model
            # also avoids FileNotFoundError on CI
        # this is basically the body of the ``config.parse.from_toml`` function.
        config_dict = {}
        for section_name in list(vak.config.parse.SECTION_CLASSES.keys()):
            if section_name in config_toml:
                vak.config.validators.are_options_valid(
                    config_toml, section_name, toml_path
                )
                section = vak.config.parse.parse_config_section(
                    config_toml, section_name, toml_path
                )
                config_dict[section_name.lower()] = section

        config = vak.config.parse.Config(**config_dict)
        assert isinstance(config, vak.config.parse.Config)



"""tests for vak.cli.predict module"""
import pytest

import vak.cli.predict
import vak.config
import vak.constants
import vak.paths

from . import cli_asserts
from ..test_core.test_predict import predict_output_matches_expected


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
    ],
)
def test_predict(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    output_dir = tmp_path.joinpath(
        f"test_predict_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "PREDICT", "option": "output_dir", "value": str(output_dir)},
        {"section": "PREDICT", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="predict",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        options_to_change=options_to_change,
    )

    vak.cli.predict.predict(toml_path)

    cfg = vak.config.parse.from_toml_path(toml_path)
    assert predict_output_matches_expected(output_dir, cfg.predict.annot_csv_filename)

    assert cli_asserts.log_file_created(command="predict", output_path=output_dir)



"""tests for vak.io.dataframe module"""
from pathlib import Path

import pandas as pd

import vak.annotation
import vak.constants
import vak.io.audio
import vak.io.dataframe
import vak.io.spect


def returned_dataframe_matches_expected(
    df_returned,
    data_dir,
    labelset,
    annot_format,
    audio_format,
    spect_format,
    spect_output_dir=None,
    annot_file=None,
    expected_audio_paths=None,
    not_expected_audio_paths=None,
    spect_file_ext=".spect.npz",
    expected_spect_paths=None,
    not_expected_spect_paths=None,
):
    """tests that dataframe returned by ``vak.io.dataframe.from_files``
    matches expected dataframe."""
    assert isinstance(df_returned, pd.DataFrame)

    assert df_returned.columns.values.tolist() == vak.io.spect.DF_COLUMNS

    annot_format_from_df = df_returned.annot_format.unique()
    assert len(annot_format_from_df) == 1
    annot_format_from_df = annot_format_from_df[0]

    if annot_format is not None:
        # test format from dataframe is specified format
        assert annot_format_from_df == annot_format
        annot_list = vak.annotation.from_df(df_returned)
    else:
        # if no annotation format specified, check that `annot_format` is `none`
        assert annot_format_from_df == vak.constants.NO_ANNOTATION_FORMAT
        annot_list = None

    if annot_file:
        assert all(
            [Path(annot_path) == annot_file for annot_path in df_returned["annot_path"]]
        )

    if labelset:
        # test that filtering vocalizations by labelset worked
        assert annot_list is not None, "labelset specified but annot_list was None"
        for annot in annot_list:
            labels = annot.seq.labels
            labelset_from_labels = set(labels)
            # should be true that set of labels from each annotation is a subset of labelset
            assert labelset_from_labels.issubset(set(labelset))

    audio_files_from_df = [Path(audio_path) for audio_path in df_returned.audio_path]
    spect_paths_from_df = [Path(spect_path) for spect_path in df_returned.spect_path]

    # --- which assertions to run depends on whether we made the dataframe
    # from audio files or spect files ----
    if audio_format:  # implies that --> we made the dataframe from audio files

        # test that all audio files came from data_dir
        for audio_file_from_df in audio_files_from_df:
            assert Path(audio_file_from_df).parent == data_dir

        # test that each audio file has a corresponding spect file in `spect_path` column
        spect_file_names = [spect_path.name for spect_path in spect_paths_from_df]
        expected_spect_files = [
            source_audio_file.name + spect_file_ext
            for source_audio_file in expected_audio_paths
        ]
        assert all(
            [
                expected_spect_file in spect_file_names
                for expected_spect_file in expected_spect_files
            ]
        )

        # if there are audio files we expect to **not** be in audio_path
        # -- because the associated annotations have labels not in labelset --
        # then test those files are **not** in spect_path
        if not_expected_audio_paths is not None:
            not_expected_spect_files = [
                source_audio_file.name + spect_file_ext
                for source_audio_file in not_expected_audio_paths
            ]
            assert all(
                [
                    not_expected_spect_file not in spect_file_names
                    for not_expected_spect_file in not_expected_spect_files
                ]
            )

        # test that all the generated spectrogram files are in a
        # newly-created directory inside spect_output_dir
        assert all(
            [
                spect_path.parents[1] == spect_output_dir
                for spect_path in spect_paths_from_df
            ]
        )

    elif spect_format:  # implies that --> we made the dataframe from audio files
        for expected_spect_path in list(expected_spect_paths):
            assert expected_spect_path in spect_paths_from_df
            spect_paths_from_df.remove(expected_spect_path)

        # test that **only** expected paths were in DataFrame
        if not_expected_spect_paths is not None:
            for not_expected_spect_path in not_expected_spect_paths:
                assert not_expected_spect_path not in spect_paths_from_df

        # test that **only** expected paths were in DataFrame
        # spect_paths_from_df should be empty after popping off all the expected paths
        assert (
            len(spect_paths_from_df) == 0
        )  # yes I know this isn't "Pythonic". It's readable, go away.

    return True  # all asserts passed


def test_from_files_with_audio_cbin_with_labelset(
    audio_dir_cbin,
    default_spect_params,
    labelset_notmat,
    audio_list_cbin_all_labels_in_labelset,
    audio_list_cbin_labels_not_in_labelset,
    spect_list_npz_all_labels_in_labelset,
    spect_list_npz_labels_not_in_labelset,
    tmp_path,
):
    """test that ``vak.io.dataframe.from_files`` works
    when we point it at directory of .cbin audio files
    and specify an annotation format"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=audio_dir_cbin,
        labelset=labelset_notmat,
        annot_format="notmat",
        audio_format="cbin",
        spect_output_dir=tmp_path,
        spect_format=None,
        annot_file=None,
        spect_params=default_spect_params,
    )

    assert returned_dataframe_matches_expected(
        vak_df,
        data_dir=audio_dir_cbin,
        labelset=labelset_notmat,
        annot_format="notmat",
        audio_format="cbin",
        spect_format=None,
        spect_output_dir=tmp_path,
        annot_file=None,
        expected_audio_paths=audio_list_cbin_all_labels_in_labelset,
        not_expected_audio_paths=audio_list_cbin_labels_not_in_labelset,
        expected_spect_paths=spect_list_npz_all_labels_in_labelset,
        not_expected_spect_paths=spect_list_npz_labels_not_in_labelset,
    )


def test_from_files_with_audio_cbin_no_annot(
    audio_dir_cbin, default_spect_params, labelset_notmat, audio_list_cbin, tmp_path
):
    """test that ``vak.io.dataframe.from_files`` works
    when we point it at directory of .cbin audio files
    and  **do not** specify an annotation format"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=audio_dir_cbin,
        annot_format=None,
        labelset=None,
        audio_format="cbin",
        spect_output_dir=tmp_path,
        spect_format=None,
        annot_file=None,
        spect_params=default_spect_params,
    )

    assert returned_dataframe_matches_expected(
        vak_df,
        data_dir=audio_dir_cbin,
        annot_format=None,
        labelset=None,
        audio_format="cbin",
        spect_format=None,
        spect_output_dir=tmp_path,
        expected_audio_paths=audio_list_cbin,
        annot_file=None,
    )


def test_from_files_with_audio_cbin_no_labelset(
    audio_dir_cbin, default_spect_params, audio_list_cbin, tmp_path
):
    """test that ``vak.io.dataframe.from_files`` works
    when we point it at directory of .cbin audio files
    and specify an annotation format"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=audio_dir_cbin,
        annot_format="notmat",
        labelset=None,
        audio_format="cbin",
        spect_format=None,
        spect_output_dir=tmp_path,
        annot_file=None,
        spect_params=default_spect_params,
    )

    assert returned_dataframe_matches_expected(
        vak_df,
        data_dir=audio_dir_cbin,
        annot_format="notmat",
        labelset=None,
        audio_format="cbin",
        spect_format=None,
        spect_output_dir=tmp_path,
        expected_audio_paths=audio_list_cbin,
        annot_file=None,
    )


def test_from_files_with_spect_mat(
    spect_dir_mat,
    labelset_yarden,
    annot_file_yarden,
    spect_list_mat_all_labels_in_labelset,
    spect_list_mat_labels_not_in_labelset,
):
    """test that ``vak.io.dataframe.from_files`` works
    when we point it at directory of .mat array files
    and specify an annotation format"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=spect_dir_mat,
        labelset=labelset_yarden,
        annot_format="yarden",
        audio_format=None,
        spect_format="mat",
        annot_file=annot_file_yarden,
        spect_params=None,
    )

    assert returned_dataframe_matches_expected(
        vak_df,
        data_dir=spect_dir_mat,
        labelset=labelset_yarden,
        annot_format="yarden",
        audio_format=None,
        spect_format="mat",
        annot_file=annot_file_yarden,
        expected_spect_paths=spect_list_mat_all_labels_in_labelset,
        not_expected_spect_paths=spect_list_mat_labels_not_in_labelset,
    )


def test_from_files_with_spect_mat_no_annot(spect_dir_mat, spect_list_mat):
    """test that ``vak.io.dataframe.from_files`` works
    when we point it at directory of .mat array files
    and **do not** specify an annotation format"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=spect_dir_mat,
        labelset=None,
        annot_format=None,
        audio_format=None,
        spect_format="mat",
        annot_file=None,
        spect_params=None,
    )

    assert returned_dataframe_matches_expected(
        vak_df,
        data_dir=spect_dir_mat,
        labelset=None,
        annot_format=None,
        audio_format=None,
        spect_format="mat",
        annot_file=None,
        expected_spect_paths=spect_list_mat,
    )


def test_from_files_with_spect_mat_no_labelset(
    spect_dir_mat, labelset_yarden, annot_file_yarden, annot_list_yarden, spect_list_mat
):
    """test that ``vak.io.dataframe.from_files`` works
    when we point it at directory of .mat array files
    and specify an annotation format
    but do not specify a labelset"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=spect_dir_mat,
        labelset=None,
        annot_format="yarden",
        audio_format=None,
        spect_format="mat",
        annot_file=annot_file_yarden,
        spect_params=None,
    )

    assert returned_dataframe_matches_expected(
        vak_df,
        data_dir=spect_dir_mat,
        labelset=None,
        annot_format="yarden",
        audio_format=None,
        spect_format="mat",
        annot_file=annot_file_yarden,
        expected_spect_paths=spect_list_mat,
    )


def test_add_split_col(audio_dir_cbin, default_spect_params, labelset_notmat, tmp_path):
    """test that ``add_split_col`` adds a 'split' column
    to a DataFrame, where all values in the Series are the
    specified split (a string)"""
    vak_df = vak.io.dataframe.from_files(
        data_dir=audio_dir_cbin,
        labelset=labelset_notmat,
        annot_format="notmat",
        audio_format="cbin",
        spect_format=None,
        annot_file=None,
        spect_params=default_spect_params,
    )

    assert "split" not in vak_df.columns

    vak_df = vak.io.dataframe.add_split_col(vak_df, split="train")
    assert "split" in vak_df.columns

    assert vak_df["split"].unique().item() == "train"



"""module that contains helper function to load metrics

Metrics in separate packages should make themselves available to vak by including
'vak.metrics' in the entry_points of their setup.py file.

For more detail on entry points in Python, see:
https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata
https://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins
https://amir.rachum.com/blog/2017/07/28/python-entry-points/
"""
from .. import entry_points

METRICS_ENTRY_POINT = "vak.models"


def find():
    """find installed vak.metrics

    returns generator that yields metric name and function for loading
    """
    for entrypoint in entry_points._iter(METRICS_ENTRY_POINT):
        yield entrypoint.name, entrypoint.load()



"""functions for plotting annotations for vocalizations"""
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
import numpy as np


def plot_segments(onsets, offsets, y=0.5, ax=None, line_kwargs=None):
    """plot segments on an axis.

    Creates a collection of horizontal lines
    with the specified `onsets` and `offsets`
    all at height `y` and places them on the axes `ax`.

    Parameters
    ----------
    onsets : numpy.ndarray
        onset times of segments
    offsets : numpy.ndarray
        offset times of segments
    y : float, int
        height on y-axis at which segments should be plotted.
        Default is 0.5.
    ax : matplotlib.axes.Axes
        axes on which to plot segment. Default is None,
        in which case a new Axes instance is created
    line_kwargs : dict
        keyword arguments passed to the `LineCollection`
        that represents the segments. Default is None.
    """
    if line_kwargs is None:
        line_kwargs = {}

    if ax is None:
        fig, ax = plt.subplots
    segments = []
    for on, off in zip(onsets, offsets):
        segments.append(((on, y), (off, y)))
    lc = LineCollection(segments, **line_kwargs)
    ax.add_collection(lc)


def plot_labels(labels, t, y=0.6, ax=None, text_kwargs=None):
    """plot labels on an axis.

    Parameters
    ----------
    labels : list, numpy.ndarray

    t : numpy.ndarray
        times at which to plot labels
    y : float, int
        height on y-axis at which segments should be plotted.
        Default is 0.5.
    ax : matplotlib.axes.Axes
        axes on which to plot segment. Default is None,
        in which case a new Axes instance is created
    text_kwargs : dict
        keyword arguments passed to the `Axes.text` method
        that plots the labels. Default is None.
    """
    if text_kwargs is None:
        text_kwargs = {}

    if ax is None:
        fig, ax = plt.subplots
    for label, t_lbl in zip(labels, t):
        ax.text(t_lbl, y, label, **text_kwargs)


def annotation(
    annot,
    t,
    tlim=None,
    y_segments=0.5,
    y_labels=0.6,
    line_kwargs=None,
    text_kwargs=None,
    ax=None,
):
    """plot segments with labels, from annotation

    Parameters
    ----------
    annot : crowsetta.Annotation
        annotation that has segments to be plotted
        (the `annot.seq.segments` attribute)
    t : numpy.ndarray
        vector of centers of time bins from spectrogram
    tlim : tuple, list
        limits of time axis (tmin, tmax) (i.e., x-axis).
        Default is None, in which case entire range of t will be plotted.
    y_segments : float
        height at which segments should be plotted.
        Default is 0.5 (assumes y-limits of 0 and 1).
    y_labels : float
        height at which labels should be plotted.
        Default is 0.6 (assumes y-limits of 0 and 1).
    line_kwargs : dict
        keyword arguments for `LineCollection`.
        Passed to the function `vak.plot.annot.segments` that plots segments
        as a `LineCollection` instance. Default is None.
    text_kwargs : dict
        keyword arguments for `matplotlib.axes.Axes.text`.
        Passed to the function `vak.plot.annot.labels` that plots labels
        using Axes.text method. Default is None.
    ax : matplotlib.axes.Axes
        axes on which to plot segments.
        Default is None, in which case
        a new figure with a single axes is created
    """
    if ax is None:
        fig, ax = plt.subplots()
        ax.set_ylim(0, 1)

    segment_centers = []
    for on, off in zip(annot.seq.onsets_s, annot.seq.offsets_s):
        segment_centers.append(np.mean([on, off]))
    plot_segments(
        onsets=annot.seq.onsets_s,
        offsets=annot.seq.offsets_s,
        y=y_segments,
        ax=ax,
        line_kwargs=line_kwargs,
    )

    if tlim:
        ax.set_xlim(tlim)
        tmin, tmax = tlim

        labels = []
        segment_centers_tmp = []
        for label, segment_center in zip(annot.seq.labels, segment_centers):
            if tmin < segment_center < tmax:
                labels.append(label)
                segment_centers_tmp.append(segment_center)
        segment_centers = segment_centers_tmp
    else:
        labels = annot.seq.labels

    segment_centers = np.array(segment_centers)
    plot_labels(
        labels=labels, t=segment_centers, y=y_labels, ax=ax, text_kwargs=text_kwargs
    )



"""tests for vak.core.train module"""
import pytest

import vak.config
import vak.constants
import vak.paths
import vak.core.train


def train_output_matches_expected(cfg, model_config_map, results_path):
    assert results_path.joinpath("labelmap.json").exists()

    if cfg.train.normalize_spectrograms or cfg.train.spect_scaler_path:
        assert results_path.joinpath("StandardizeSpect").exists()
    else:
        assert not results_path.joinpath("StandardizeSpect").exists()

    for model_name in model_config_map.keys():
        model_path = results_path.joinpath(model_name)
        assert model_path.exists()

        tensorboard_log = sorted(model_path.glob(f"events.out.tfevents.*{model_name}"))
        assert len(tensorboard_log) == 1

        checkpoints_path = model_path.joinpath("checkpoints")
        assert checkpoints_path.exists()
        assert checkpoints_path.joinpath("checkpoint.pt").exists()
        if cfg.train.val_step is not None:
            assert checkpoints_path.joinpath("max-val-acc-checkpoint.pt").exists()

    return True


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_train(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    options_to_change = {"section": "TRAIN", "option": "device", "value": device}
    toml_path = specific_config(
        config_type="train",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    vak.core.train(
        model_config_map,
        cfg.train.csv_path,
        cfg.dataloader.window_size,
        cfg.train.batch_size,
        cfg.train.num_epochs,
        cfg.train.num_workers,
        labelset=cfg.prep.labelset,
        results_path=results_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.train.normalize_spectrograms,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
        patience=cfg.train.patience,
        device=cfg.train.device,
        logger=None,
    )

    assert train_output_matches_expected(cfg, model_config_map, results_path)


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_continue_training(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    options_to_change = {"section": "TRAIN", "option": "device", "value": device}
    toml_path = specific_config(
        config_type="train_continue",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    vak.core.train(
        model_config_map=model_config_map,
        csv_path=cfg.train.csv_path,
        window_size=cfg.dataloader.window_size,
        batch_size=cfg.train.batch_size,
        num_epochs=cfg.train.num_epochs,
        num_workers=cfg.train.num_workers,
        labelmap_path=cfg.train.labelmap_path,
        spect_scaler_path=cfg.train.spect_scaler_path,
        results_path=results_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.train.normalize_spectrograms,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
        patience=cfg.train.patience,
        device=cfg.train.device,
        logger=None,
    )

    assert train_output_matches_expected(cfg, model_config_map, results_path)



"""parses [PREDICT] section of config"""
import os
from pathlib import Path

import attr
from attr import converters, validators
from attr.validators import instance_of

from .validators import is_a_directory, is_a_file, is_valid_model_name
from .. import device
from ..converters import comma_separated_list, expanded_user_path


@attr.s
class PredictConfig:
    """class that represents [PREDICT] section of config.toml file

     Attributes
     ----------
     csv_path : str
         path to where dataset was saved as a csv.
     checkpoint_path : str
         path to directory with checkpoint files saved by Torch, to reload model
     labelmap_path : str
         path to 'labelmap.json' file.
     models : list
         of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
     batch_size : int
         number of samples per batch presented to models during training.
     num_workers : int
         Number of processes to use for parallel loading of data.
         Argument to torch.DataLoader. Default is 2.
     device : str
         Device on which to work with model + data.
         Defaults to 'cuda' if torch.cuda.is_available is True.
     spect_scaler_path : str
         path to a saved SpectScaler object used to normalize spectrograms.
         If spectrograms were normalized and this is not provided, will give
         incorrect results.
     annot_csv_filename : str
         name of .csv file containing predicted annotations.
         Default is None, in which case the name of the dataset .csv
         is used, with '.annot.csv' appended to it.
     output_dir : str
         path to location where .csv containing predicted annotation
         should be saved. Defaults to current working directory.
     min_segment_dur : float
         minimum duration of segment, in seconds. If specified, then
         any segment with a duration less than min_segment_dur is
         removed from lbl_tb. Default is None, in which case no
         segments are removed.
     majority_vote : bool
         if True, transform segments containing multiple labels
         into segments with a single label by taking a "majority vote",
         i.e. assign all time bins in the segment the most frequently
         occurring label in the segment. This transform can only be
         applied if the labelmap contains an 'unlabeled' label,
         because unlabeled segments makes it possible to identify
         the labeled segments. Default is False.
    save_net_outputs : bool
         if True, save 'raw' outputs of neural networks
         before they are converted to annotations. Default is False.
         Typically the output will be "logits"
         to which a softmax transform might be applied.
         For each item in the dataset--each row in  the `csv_path` .csv--
         the output will be saved in a separate file in `output_dir`,
         with the extension `{MODEL_NAME}.output.npz`. E.g., if the input is a
         spectrogram with `spect_path` filename `gy6or6_032312_081416.npz`,
         and the network is `TweetyNet`, then the net output file
         will be `gy6or6_032312_081416.tweetynet.output.npz`.
    """

    # required, external files
    checkpoint_path = attr.ib(converter=expanded_user_path, validator=is_a_file)
    labelmap_path = attr.ib(converter=expanded_user_path, validator=is_a_file)

    # required, model / dataloader
    models = attr.ib(
        converter=comma_separated_list,
        validator=[instance_of(list), is_valid_model_name],
    )
    batch_size = attr.ib(converter=int, validator=instance_of(int))

    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )

    # optional, transform
    spect_scaler_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )

    # optional, data loader
    num_workers = attr.ib(validator=instance_of(int), default=2)
    device = attr.ib(validator=instance_of(str), default=device.get_default())

    annot_csv_filename = attr.ib(
        validator=validators.optional(instance_of(str)), default=None
    )
    output_dir = attr.ib(
        converter=expanded_user_path,
        validator=is_a_directory,
        default=Path(os.getcwd()),
    )
    min_segment_dur = attr.ib(
        validator=validators.optional(instance_of(float)), default=None
    )
    majority_vote = attr.ib(validator=instance_of(bool), default=True)
    save_net_outputs = attr.ib(validator=instance_of(bool), default=False)



import sys


def test_installed():
    # makes sure the entry point loads properly
    if "vak" in sys.modules:
        sys.modules.pop("vak")
    import vak

    models = [name for name, class_ in sorted(vak.models.find())]
    assert "TeenyTweetyNetModel" in models



import torch
from torch import nn

from ..engine.model import Model

# absolute import to avoid name clash in model def below
import vak.metrics


class TeenyTweetyNet(nn.Module):
    def __init__(
        self,
        num_classes,
        input_shape=(1, 513, 88),
        conv1_filters=8,
        conv1_kernel_size=(5, 5),
        conv1_padding=(0, 2),
        conv2_filters=16,
        conv2_kernel_size=(5, 5),
        conv2_padding=(0, 2),
        pool1_size=(8, 1),
        pool1_stride=(8, 1),
        pool2_size=(8, 1),
        pool2_stride=(8, 1),
        hidden_size=64,
    ):
        """TeenyTweetyNet model

        Parameters
        ----------
        num_classes : int
            number of classes to predict, e.g., number of syllable classes in an individual bird's song
        input_shape : tuple
            with 3 elements corresponding to dimensions of spectrogram windows: (channels, frequency bins, time bins).
            i.e. we assume input is a spectrogram and treat it like an image, typically with one channel,
            the rows are frequency bins, and the columns are time bins. Default is (1, 513, 88).
        conv1_filters : int
            Number of filters in first convolutional layer. Default is 32.
        conv1_kernel_size : tuple
            Size of kernels, i.e. filters, in first convolutional layer. Default is (5, 5).
        conv2_filters : int
            Number of filters in second convolutional layer. Default is 64.
        conv2_kernel_size : tuple
            Size of kernels, i.e. filters, in second convolutional layer. Default is (5, 5).
        pool1_size : two element tuple of ints
            Size of sliding window for first max pooling layer. Default is (1, 8)
        pool1_stride : two element tuple of ints
            Step size for sliding window of first max pooling layer. Default is (1, 8)
        pool2_size : two element tuple of ints
            Size of sliding window for second max pooling layer. Default is (1, 8),
        pool2_stride : two element tuple of ints
            Step size for sliding window of second max pooling layer. Default is (1, 8)
        """
        super().__init__()
        self.num_classes = num_classes
        self.input_shape = input_shape
        self.hidden_size = hidden_size

        self.cnn = nn.Sequential(
            nn.Conv2d(
                in_channels=self.input_shape[0],
                out_channels=conv1_filters,
                kernel_size=conv1_kernel_size,
                padding=conv1_padding,
            ),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=pool1_size, stride=pool1_stride),
            nn.Conv2d(
                in_channels=conv1_filters,
                out_channels=conv2_filters,
                kernel_size=conv2_kernel_size,
                padding=conv2_padding,
            ),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=pool2_size, stride=pool2_stride),
        )

        # determine number of features in output after stacking channels
        # we use the same number of features for hidden states
        # note self.num_hidden is also used to reshape output of cnn in self.forward method
        batch_shape = tuple((1,) + input_shape)
        tmp_tensor = torch.rand(batch_shape)
        tmp_out = self.cnn(tmp_tensor)
        channels_out, freqbins_out = tmp_out.shape[1], tmp_out.shape[2]
        self.num_rnn_features = channels_out * freqbins_out

        self.rnn = nn.LSTM(
            input_size=self.num_rnn_features,
            hidden_size=self.hidden_size,
            num_layers=1,
            dropout=0,
            bidirectional=True,
        )

        # for self.fc, in_features = hidden_size * 2, because LSTM is bidirectional
        # so we get hidden forward + hidden backward as output
        self.fc = nn.Linear(self.hidden_size * 2, num_classes)

    def forward(self, x):
        features = self.cnn(x)
        # stack channels so that dimension order is (batch, num_rnn_features, num time bins)
        features = features.view(features.shape[0], self.num_rnn_features, -1)
        # now switch dimensions for feeding to rnn,
        # so dimension order is (num time bins, batch size, num_rnn_features)
        features = features.permute(2, 0, 1)
        rnn_output, (hidden, cell_state) = self.rnn(features)
        # permute back to (batch, time bins, features)
        # so we can project features down onto number of classes
        rnn_output = rnn_output.permute(1, 0, 2)
        logits = self.fc(rnn_output)
        # permute yet again
        # so that dimension order is (batch, classes, time steps)
        # because this is order that loss function expects
        return logits.permute(0, 2, 1)


class TeenyTweetyNetModel(Model):
    @classmethod
    def from_config(cls, config, logger=None):
        network = TeenyTweetyNet(**config["network"])
        loss = nn.CrossEntropyLoss(**config["loss"])
        optimizer = torch.optim.Adam(params=network.parameters(), **config["optimizer"])
        metrics = {
            "acc": vak.metrics.Accuracy(),
            "levenshtein": vak.metrics.Levenshtein(),
            "segment_error_rate": vak.metrics.SegmentErrorRate(),
            "loss": torch.nn.CrossEntropyLoss(),
        }
        return cls(
            network=network,
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            logger=logger,
        )



"""tests for vak.config.train module"""
import vak.config.train


def test_train_attrs_class(all_generated_train_configs_toml_path_pairs):
    """test that instantiating TrainConfig class works as expected"""
    for config_toml, toml_path in all_generated_train_configs_toml_path_pairs:
        train_section = config_toml["TRAIN"]
        train_config_obj = vak.config.train.TrainConfig(**train_section)
        assert isinstance(train_config_obj, vak.config.train.TrainConfig)



import numpy as np
import torch

__all__ = [
    "pad_to_window",
    "standardize_spect",
    "to_floattensor",
    "to_longtensor",
    "view_as_window_batch",
]


def standardize_spect(spect, mean_freqs, std_freqs, non_zero_std):
    """standardize spectrogram by subtracting off mean and dividing by standard deviation.

    Parameters
    ----------
    spect : numpy.ndarray
        with shape (frequencies, time bins)
    mean_freqs : numpy.ndarray
        vector of mean values for each frequency bin across the fit set of spectrograms
    std_freqs : numpy.ndarray
        vector of standard deviations for each frequency bin across the fit set of spectrograms
    non_zero_std : numpy.ndarray
        boolean, indicates where std_freqs has non-zero values. Used to avoid divide-by-zero errors.

    Returns
    -------
    transformed : numpy.ndarray
        with same shape as spect but with (approximately) zero mean and unit standard deviation
        (mean and standard devation will still vary by batch).
    """
    tfm = spect - mean_freqs[:, np.newaxis]  # need axis for broadcasting
    # keep any stds that are zero from causing NaNs
    tfm[non_zero_std, :] = tfm[non_zero_std, :] / std_freqs[non_zero_std, np.newaxis]
    return tfm


def pad_to_window(arr, window_size, padval=0.0, return_padding_mask=True):
    """pad a 1d or 2d array so that it can be reshaped
    into consecutive windows of specified size

    Parameters
    ----------
    arr : numpy.ndarray
        with 1 or 2 dimensions, e.g. a vector of labeled timebins
        or a spectrogram.
    window_size : int
        width of window in number of elements.
    padval : float
        value to pad with. Added to end of array, the
        "right side" if 2-dimensional.
    return_padding_mask : bool
        if True, return a boolean vector to use for cropping
        back down to size before padding. padding_mask has size
        equal to width of padded array, i.e. original size
        plus padding at the end, and has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.

    Returns
    -------
    padded : numpy.ndarray
        padded with padval
    padding_mask : np.bool
        has size equal to width of padded, i.e. original size
        plus padding at the end. Has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.
        Only returned if return_padding_mask is True.
    """
    if arr.ndim == 1:
        width = arr.shape[0]
    elif arr.ndim == 2:
        height, width = arr.shape
    else:
        raise ValueError(
            f"input array must be 1d or 2d but number of dimensions was: {arr.ndim}"
        )

    target_width = int(np.ceil(width / window_size) * window_size)

    if arr.ndim == 1:
        padded = np.ones((target_width,)) * padval
        padded[:width] = arr
    elif arr.ndim == 2:
        padded = np.ones((height, target_width)) * padval
        padded[:, :width] = arr

    if return_padding_mask:
        padding_mask = np.zeros((target_width,), dtype=np.bool)
        padding_mask[:width] = True
        return padded, padding_mask
    else:
        return padded


def view_as_window_batch(arr, window_width):
    """return view of a 1d or 2d array as a batch of non-overlapping windows

    Parameters
    ----------
    arr : numpy.ndarray
        with 1 or 2 dimensions, e.g. a vector of labeled timebins
        or a 2-d array representing a spectrogram.
        If the array has 2-d dimensions, the returned array will
        have dimensions (batch, height of array, window width)
    window_width : int
        width of window in number of elements.

    Returns
    -------
    batch_windows : numpy.ndarray
        with shape (batch size, window_size) if array is 1d,
        or with shape (batch size, height, window_size) if array is 2d.
        Batch size will be arr.shape[-1] // window_width.
        Window width must divide arr.shape[-1] evenly.
        To pad the array so it can be divided into windows of the specified
        width, use the `pad_to_window` transform

    Notes
    -----
    adapted from skimage.util.view_as_blocks
    https://github.com/scikit-image/scikit-image/blob/f1b7cf60fb80822849129cb76269b75b8ef18db1/skimage/util/shape.py#L9
    """
    if not (type(window_width) == int and window_width > 0):
        raise ValueError(f"window width must be a positive integer")

    if arr.ndim == 1:
        window_shape = (window_width,)
    elif arr.ndim == 2:
        height, _ = arr.shape
        window_shape = (height, window_width)
    else:
        raise ValueError(
            f"input array must be 1d or 2d but number of dimensions was: {arr.ndim}"
        )

    window_shape = np.array(window_shape)
    arr_shape = np.array(arr.shape)
    if (arr_shape % window_shape).sum() != 0:
        raise ValueError(
            "'window_width' does not divide evenly into with 'arr' shape. "
            "Use 'pad_to_window' transform to pad array so it can be windowed."
        )

    new_shape = tuple(arr_shape // window_shape) + tuple(window_shape)
    new_strides = tuple(arr.strides * window_shape) + arr.strides
    batch_windows = np.lib.stride_tricks.as_strided(
        arr, shape=new_shape, strides=new_strides
    )
    # when 2d, first dim 1 because new shape has height equal to original arr
    if batch_windows.ndim == 4 and batch_windows.shape[0] == 1:
        # we don't want that extra dim of size 1, we want first dim to be "batch"
        batch_windows = np.squeeze(batch_windows)
    return batch_windows


def to_floattensor(arr):
    """convert Numpy array to torch.FloatTensor.

    Parameters
    ----------
    arr : numpy.ndarray

    Returns
    -------
    float_tensor
        with dtype 'float32'
    """
    return torch.from_numpy(arr).float()


def to_longtensor(arr):
    """convert Numpy array to torch.LongTensor.

    Parameters
    ----------
    arr : numpy.ndarray

    Returns
    -------
    long_tensor : torch.Tensor
        with dtype 'float64'
    """
    return torch.from_numpy(arr).long()


def add_channel(input, channel_dim=0):
    """add a channel dimension to a tensor.
    Transform that makes it easy to treat a spectrogram as an image,
    by adding a dimension with a single 'channel', analogous to grayscale.
    In this way the tensor can be fed to e.g. convolutional layers.

    Parameters
    ----------
    input : torch.Tensor
    channel_dim : int
        dimension where "channel" is added. Default is 0.
    """
    return torch.unsqueeze(input, dim=channel_dim)



from collections import OrderedDict
from datetime import datetime
import json

import joblib
import pandas as pd
import torch.utils.data

from .. import models
from .. import transforms
from ..datasets.vocal_dataset import VocalDataset
from ..logging import log_or_print


def eval(
    csv_path,
    model_config_map,
    checkpoint_path,
    labelmap_path,
    output_dir,
    window_size,
    num_workers,
    split="test",
    spect_scaler_path=None,
    spect_key="s",
    timebins_key="t",
    device=None,
    logger=None,
):
    """evaluate a trained model

    Parameters
    ----------
    csv_path : str, pathlib.Path
        path to where dataset was saved as a csv.
    model_config_map : dict
        where each key-value pair is model name : dict of config parameters
    checkpoint_path : str, pathlib.Path
        path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str, pathlib.Path
        Path to location where .csv files with evaluation metrics should be saved.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shown to neural networks
    labelmap_path : str, pathlib.Path
        path to 'labelmap.json' file.
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    batch_size : int
        number of samples per batch presented to models during training.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    split : str
        split of dataset on which model should be evaluated.
        One of {'train', 'val', 'test'}. Default is 'test'.
    spect_scaler_path : str, pathlib.Path
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
        Default is None.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    None
    """
    # ---- get time for .csv file --------------------------------------------------------------------------
    timenow = datetime.now().strftime("%y%m%d_%H%M%S")

    # ---------------- load data for evaluation ------------------------------------------------------------------------
    if spect_scaler_path:
        log_or_print(
            f"loading spect scaler from path: {spect_scaler_path}",
            logger=logger,
            level="info",
        )
        spect_standardizer = joblib.load(spect_scaler_path)
    else:
        log_or_print(
            f"not using a spect scaler",
            logger=logger,
            level="info",
        )
        spect_standardizer = None

    log_or_print(
        f"loading labelmap from path: {labelmap_path}", logger=logger, level="info"
    )
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)

    item_transform = transforms.get_defaults(
        "eval",
        spect_standardizer,
        window_size=window_size,
        return_padding_mask=True,
    )
    log_or_print(
        f"creating dataset for evaluation from: {csv_path}",
        logger=logger,
        level="info",
    )
    val_dataset = VocalDataset.from_csv(
        csv_path=csv_path,
        split=split,
        labelmap=labelmap,
        spect_key=spect_key,
        timebins_key=timebins_key,
        item_transform=item_transform,
    )
    val_data = torch.utils.data.DataLoader(
        dataset=val_dataset,
        shuffle=False,
        # batch size 1 because each spectrogram reshaped into a batch of windows
        batch_size=1,
        num_workers=num_workers,
    )

    # ---------------- do the actual evaluating ------------------------------------------------------------------------
    input_shape = val_dataset.shape
    # if dataset returns spectrogram reshaped into windows,
    # throw out the window dimension; just want to tell network (channels, height, width) shape
    if len(input_shape) == 4:
        input_shape = input_shape[1:]

    models_map = models.from_model_config_map(
        model_config_map, num_classes=len(labelmap), input_shape=input_shape
    )

    for model_name, model in models_map.items():
        log_or_print(
            f"running evaluation for model: {model_name}", logger=logger, level="info"
        )
        model.load(checkpoint_path, device=device)
        metric_vals = model.evaluate(eval_data=val_data, device=device)
        # create a "DataFrame" with just one row which we will save as a csv;
        # the idea is to be able to concatenate csvs from multiple runs of eval
        row = OrderedDict(
            [
                ("model_name", model_name),
                ("checkpoint_path", checkpoint_path),
                ("labelmap_path", labelmap_path),
                ("spect_scaler_path", spect_scaler_path),
                ("csv_path", csv_path),
            ]
        )
        # order metrics by name to be extra sure they will be consistent across runs
        row.update(
            sorted([(k, v) for k, v in metric_vals.items() if k.startswith("avg_")])
        )

        # pass index into dataframe, needed when using all scalar values (a single row)
        # throw away index below when saving to avoid extra column
        eval_df = pd.DataFrame(row, index=[0])
        eval_csv_path = output_dir.joinpath(f"eval_{model_name}_{timenow}.csv")
        log_or_print(
            f"saving csv with evaluation metrics at: {eval_csv_path}",
            logger=logger,
            level="info",
        )
        eval_df.to_csv(
            eval_csv_path, index=False
        )  # index is False to avoid having "Unnamed: 0" column when loading



import pandas as pd
from torch.utils.tensorboard import SummaryWriter

import vak.tensorboard


FILENAME_SUFFIX = "FakeModelName"


def test_summary_writer(tmp_path):
    writer = vak.tensorboard.get_summary_writer(
        log_dir=str(tmp_path), filename_suffix=FILENAME_SUFFIX
    )
    assert isinstance(writer, SummaryWriter)


def test_events2df(events_path):
    df = vak.tensorboard.events2df(events_path)

    assert isinstance(df, pd.DataFrame)
    assert df.index.name == "step"



import vak.files.spect
import vak.labels


def test_to_map():
    labelset = set(list("abcde"))
    labelmap = vak.labels.to_map(labelset, map_unlabeled=False)
    assert type(labelmap) == dict
    assert len(labelmap) == len(labelset)  # because map_unlabeled=False

    labelset = set(list("abcde"))
    labelmap = vak.labels.to_map(labelset, map_unlabeled=True)
    assert type(labelmap) == dict
    assert len(labelmap) == len(labelset) + 1  # because map_unlabeled=True

    labelset = {1, 2, 3, 4, 5, 6}
    labelmap = vak.labels.to_map(labelset, map_unlabeled=False)
    assert type(labelmap) == dict
    assert len(labelmap) == len(labelset)  # because map_unlabeled=False

    labelset = {1, 2, 3, 4, 5, 6}
    labelmap = vak.labels.to_map(labelset, map_unlabeled=True)
    assert type(labelmap) == dict
    assert len(labelmap) == len(labelset) + 1  # because map_unlabeled=True


def test_to_set():
    labels1 = [1, 1, 1, 1, 2, 2, 3, 3, 3]
    labels2 = [1, 1, 1, 2, 2, 3, 3, 3, 3, 3]
    labels_list = [labels1, labels2]
    labelset = vak.labels.to_set(labels_list)
    assert type(labelset) == set
    assert labelset == {1, 2, 3}



"""constants used by multiple modules.
Defined here to avoid circular imports.
"""
from functools import partial

import crowsetta
import numpy as np
from evfuncs import load_cbin
from scipy.io import loadmat
import soundfile


AUDIO_FORMAT_FUNC_MAP = {
    "cbin": load_cbin,
    "flac": soundfile.read,
    "ogg": soundfile.read,
    "wav": soundfile.read,
}

VALID_AUDIO_FORMATS = list(AUDIO_FORMAT_FUNC_MAP.keys())

# ---- spectrogram files ----
SPECT_FORMAT_LOAD_FUNCTION_MAP = {
    "mat": partial(loadmat, squeeze_me=True),
    "npz": np.load,
}
VALID_SPECT_FORMATS = list(SPECT_FORMAT_LOAD_FUNCTION_MAP.keys())

# ---- annotation files ----
VALID_ANNOT_FORMATS = crowsetta.formats._INSTALLED
NO_ANNOTATION_FORMAT = "none"

# format for timestamps
STRFTIME_TIMESTAMP = "%y%m%d_%H%M%S"

# ---- results, from train / learncurve ----
RESULTS_DIR_PREFIX = "results_"

# ---- output (default) file extensions. Using the `pathlib` name "suffix" ----
ANNOT_CSV_SUFFIX = ".annot.csv"
NET_OUTPUT_SUFFIX = ".output.npz"



from pathlib import Path

import pandas as pd
import pytest


@pytest.fixture
def specific_csv_path(specific_config_toml):
    """returns a function that will return the
    ``csv_path`` corresponding to the ``prep``ared dataset
    from a specific configuration file,
    determined by characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`
    """

    def _specific_csv_path(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
    ):
        config_toml = specific_config_toml(
            config_type, model, annot_format, audio_format, spect_format
        )
        return Path(config_toml[config_type.upper()]["csv_path"])
        _return_toml(config_path)

    return _specific_csv_path


@pytest.fixture
def specific_dataframe(specific_csv_path):
    """returns a function that will return a
    dataframe corresponding to the ``prep``ared dataset
    from a specific configuration file,
    determined by characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`
    """

    def _specific_dataframe(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
    ):
        csv_path = specific_csv_path(
            config_type, model, annot_format, audio_format, spect_format
        )
        return pd.read_csv(csv_path)

    return _specific_dataframe



from typing import Optional

import torch


__all__ = ["one_hot"]


# adapted from kornia, https://github.com/kornia/kornia/blob/master/kornia/utils/one_hot.py
def one_hot(
    labels: torch.Tensor,
    num_classes: int,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
    eps: float = 1e-6,
) -> torch.Tensor:
    r"""Converts an integer label x-D tensor to a one-hot (x+1)-D tensor.
    Args:
        labels (torch.Tensor) : tensor with labels of shape :math:`(N, *)`,
                                where N is batch size. Each value is an integer
                                representing correct classification.
        num_classes (int): number of classes in labels.
        device (Optional[torch.device]): the desired device of returned tensor.
         Default: if None, defaults to device of `labels`.
        dtype (Optional[torch.dtype]): the desired data type of returned
         tensor. Default: if None, infers data type from values.
    Returns:
        torch.Tensor: the labels in one hot tensor of shape :math:`(N, C, *)`,
    Examples:
        >>> labels = torch.LongTensor([[[0, 1], [2, 0]]])
        >>> one_hot(labels, num_classes=3)
        tensor([[[[1.0000e+00, 1.0000e-06],
                  [1.0000e-06, 1.0000e+00]],
        <BLANKLINE>
                 [[1.0000e-06, 1.0000e+00],
                  [1.0000e-06, 1.0000e-06]],
        <BLANKLINE>
                 [[1.0000e-06, 1.0000e-06],
                  [1.0000e+00, 1.0000e-06]]]])
    """
    if not isinstance(labels, torch.Tensor):
        raise TypeError(
            "Input labels type is not a torch.Tensor. Got {}".format(type(labels))
        )

    if not labels.dtype == torch.int64:
        raise ValueError(
            "labels must be of the same dtype torch.int64. Got: {}".format(labels.dtype)
        )

    if num_classes < 1:
        raise ValueError(
            "The number of classes must be bigger than one."
            " Got: {}".format(num_classes)
        )

    shape = labels.shape
    if device is None:
        device = labels.device
    one_hot = torch.zeros(
        (shape[0], num_classes) + shape[1:], device=device, dtype=dtype
    )

    return one_hot.scatter_(1, labels.unsqueeze(1), 1.0) + eps



"""tests for vak.config.parse module"""
import copy

import pytest

import vak.config
import vak.transforms.transforms
import vak.split
import vak.models
import vak.spect


@pytest.mark.parametrize(
    "section_name",
    [
        "DATALOADER",
        "EVAL" "LEARNCURVE",
        "PREDICT",
        "PREP",
        "SPECT_PARAMS",
        "TRAIN",
    ],
)
def test_parse_config_section_returns_attrs_class(
    section_name,
    all_generated_configs_toml_path_pairs,
    default_model,
):
    """test that ``vak.config.parse.parse_config_section``
    returns an instance of ``vak.config.learncurve.LearncurveConfig``"""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if default_model not in str(toml_path):
            continue  # only need to check configs for one model
            # also avoids FileNotFoundError on CI
        if section_name in config_toml:
            config_section_obj = vak.config.parse.parse_config_section(
                config_toml=config_toml,
                section_name=section_name,
                toml_path=toml_path,
            )
            assert isinstance(
                config_section_obj, vak.config.parse.SECTION_CLASSES[section_name]
            )


@pytest.mark.parametrize(
    "section_name",
    [
        "DATALOADER",
        "EVAL",
        "LEARNCURVE",
        "PREDICT",
        "PREP",
        "SPECT_PARAMS",
        "TRAIN",
    ],
)
def test_parse_config_section_missing_options_raises(
    section_name,
    all_generated_configs_toml_path_pairs,
    default_model,
):
    """test that configs without the required options in a section raise KeyError"""
    if vak.config.parse.REQUIRED_OPTIONS[section_name] is None:
        pytest.skip(f"no required options to test for section: {section_name}")

    # in comprehensions below, filter by default model
    # because we only need to check configs for one model
    # also avoids FileNotFoundError on CI
    if section_name == "PREP":
        configs_toml_path_pairs = (
            (config_toml, toml_path)
            for config_toml, toml_path in all_generated_configs_toml_path_pairs
            if default_model in str(toml_path)
        )
    else:
        configs_toml_path_pairs = (
            (config_toml, toml_path)
            for config_toml, toml_path in all_generated_configs_toml_path_pairs
            if section_name.lower() in toml_path.name
            and default_model in str(toml_path)
        )

    for config_toml, toml_path in configs_toml_path_pairs:
        if section_name in config_toml:
            for option in vak.config.parse.REQUIRED_OPTIONS[section_name]:
                config_copy = copy.deepcopy(config_toml)
                config_copy[section_name].pop(option)
                with pytest.raises(KeyError):
                    config = vak.config.parse.parse_config_section(
                        config_toml=config_copy,
                        section_name=section_name,
                        toml_path=toml_path,
                    )


@pytest.mark.parametrize("section_name", ["EVAL", "LEARNCURVE", "PREDICT", "TRAIN"])
def test_parse_config_section_model_not_installed_raises(
    section_name, all_generated_configs_toml_path_pairs
):
    """test that a ValueError is raised when the ``models`` option
    in the section specifies names of models that are not installed"""
    # we only need one toml, path pair
    # so we just call next on the ``zipped`` iterator that our fixture gives us
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these

    config_toml[section_name]["models"] = "NotInstalledNet, OtherNotInstalledNet"
    with pytest.raises(ValueError):
        vak.config.parse.parse_config_section(
            config_toml=config_toml, section_name=section_name, toml_path=toml_path
        )


@pytest.mark.parametrize(
    "section_name",
    [
        "EVAL",
        "PREDICT",
    ],
)
def test_parse_config_section_nonexistent_checkpoint_path_raises(
    section_name, all_generated_configs_toml_path_pairs
):
    """test that a FileNotFoundError is raised when checkpoint_path does not exist"""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these

    config_toml[section_name]["checkpoint_path"] = "obviously/non/existent/dir/check.pt"
    with pytest.raises(FileNotFoundError):
        vak.config.parse.parse_config_section(
            config_toml=config_toml, section_name=section_name, toml_path=toml_path
        )


@pytest.mark.parametrize("section_name", ["LEARNCURVE", "TRAIN"])
def test_parse_config_section_nonexistent_root_results_dir_raises(
    section_name, all_generated_configs_toml_path_pairs
):
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these

    config_toml[section_name]["root_results_dir"] = "obviously/non/existent/dir"
    with pytest.raises(NotADirectoryError):
        vak.config.parse.parse_config_section(config_toml, section_name, toml_path)


@pytest.mark.parametrize("section_name", ["EVAL", "LEARNCURVE", "PREDICT", "TRAIN"])
def test_nonexistent_csv_path_raises(
    section_name, all_generated_configs_toml_path_pairs
):
    """test that a FileNotFoundError is raised when csv_path does not exist"""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these

    config_toml[section_name]["csv_path"] = "obviously/non/existent/dir/predict.csv"
    with pytest.raises(FileNotFoundError):
        vak.config.parse.parse_config_section(
            config_toml=config_toml, section_name=section_name, toml_path=toml_path
        )


def test_parse_prep_section_nonexistent_data_dir_raises_error(
    all_generated_configs_toml_path_pairs,
):
    config_toml, toml_path = next(all_generated_configs_toml_path_pairs)
    config_toml["PREP"]["data_dir"] = "theres/no/way/this/is/a/dir"
    with pytest.raises(NotADirectoryError):
        vak.config.parse.parse_config_section(config_toml, "PREP", toml_path)


def test_parse_prep_section_both_audio_and_spect_format_raises(
    all_generated_configs_toml_path_pairs,
):
    """test that a config with both an audio and a spect format raises a ValueError"""
    # iterate through configs til we find one with an `audio_format` option
    # and then we'll add a `spect_format` option to it
    found_config_to_use = False
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if "audio_format" in config_toml["PREP"]:
            found_config_to_use = True
            break
    assert found_config_to_use  # sanity check

    config_toml["PREP"]["spect_format"] = "mat"
    with pytest.raises(ValueError):
        vak.config.parse.parse_config_section(config_toml, "PREP", toml_path)


def test_parse_prep_section_neither_audio_nor_spect_format_raises(
    all_generated_configs_toml_path_pairs,
):
    """test that a config without either an audio or a spect format raises a ValueError"""
    # iterate through configs til we find one with an `audio_format` option
    # and then we'll add a `spect_format` option to it
    found_config_to_use = False
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if "audio_format" in config_toml["PREP"]:
            found_config_to_use = True
            break
    assert found_config_to_use  # sanity check

    config_toml["PREP"].pop("audio_format")
    if "spect_format" in config_toml["PREP"]:
        # shouldn't be but humor me
        config_toml["PREP"].pop("spect_format")

    with pytest.raises(ValueError):
        vak.config.parse.parse_config_section(config_toml, "PREP", toml_path)


def test_load_from_toml_path(all_generated_configs):
    for toml_path in all_generated_configs:
        config_toml = vak.config.parse._load_toml_from_path(toml_path)
        assert isinstance(config_toml, dict)


def test_load_from_toml_path_raises_when_config_doesnt_exist(config_that_doesnt_exist):
    with pytest.raises(FileNotFoundError):
        vak.config.parse._load_toml_from_path(config_that_doesnt_exist)


def test_from_toml_path_returns_instance_of_config(
    all_generated_configs, default_model
):
    for toml_path in all_generated_configs:
        if default_model not in str(toml_path):
            continue  # only need to check configs for one model
            # also avoids FileNotFoundError on CI
        config_obj = vak.config.parse.from_toml_path(toml_path)
        assert isinstance(config_obj, vak.config.parse.Config)


def test_from_toml_path_raises_when_config_doesnt_exist(config_that_doesnt_exist):
    with pytest.raises(FileNotFoundError):
        vak.config.parse.from_toml_path(config_that_doesnt_exist)


def test_from_toml(all_generated_configs_toml_path_pairs, default_model):
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if default_model not in str(toml_path):
            continue  # only need to check configs for one model
            # also avoids FileNotFoundError on CI
        config_obj = vak.config.parse.from_toml(config_toml, toml_path)
        assert isinstance(config_obj, vak.config.parse.Config)


def test_from_toml_parse_prep_with_sections_not_none(
    all_generated_configs_toml_path_pairs,
):
    """test that we get only the sections we want when we pass in a sections list to
    ``from_toml``. Specifically test ``PREP`` since that's what this will be used for."""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        config_obj = vak.config.parse.from_toml(
            config_toml, toml_path, sections=["PREP", "SPECT_PARAMS"]
        )
        assert isinstance(config_obj, vak.config.parse.Config)
        for should_have in ("prep", "spect_params"):
            assert hasattr(config_obj, should_have)
        for should_be_none in ("eval", "learncurve", "train", "predict"):
            assert getattr(config_obj, should_be_none) is None
        assert (
            getattr(config_obj, "dataloader")
            == vak.config.dataloader.DataLoaderConfig()
        )


@pytest.mark.parametrize("section_name", ["EVAL", "LEARNCURVE", "PREDICT", "TRAIN"])
def test_from_toml_parse_prep_with_sections_not_none_does_not_raise(
    section_name, all_generated_configs_toml_path_pairs, random_path_factory
):
    """test that invalid values in other sections are ignored when we pass in a sections list
    specifying that we only want to parse ``PREP`` and ``SPECT_PARAMS``, e.g., ignore non-existent
    csv_path or checkpoint_path"""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these

    purpose = vak.cli.prep.purpose_from_toml(config_toml, toml_path)
    section_name = purpose.upper()
    required_options = vak.config.parse.REQUIRED_OPTIONS[section_name]
    for required_option in required_options:
        # set option to values that **would** cause an error if we parse them
        if "path" in required_option:
            badval = random_path_factory(f"_{required_option}.exe")
        elif "dir" in required_option:
            badval = random_path_factory("nonexistent_dir")
        else:
            continue
        config_toml[section_name][required_option] = badval
    config_obj = vak.config.parse.from_toml(
        config_toml, toml_path, sections=["PREP", "SPECT_PARAMS"]
    )
    with pytest.raises((FileNotFoundError, NotADirectoryError)):
        vak.config.parse.from_toml(config_toml, toml_path, sections=section_name)


def test_invalid_section_raises(invalid_section_config_path):
    with pytest.raises(ValueError):
        vak.config.parse.from_toml_path(invalid_section_config_path)


def test_invalid_option_raises(invalid_option_config_path):
    with pytest.raises(ValueError):
        vak.config.parse.from_toml_path(invalid_option_config_path)


@pytest.fixture
def invalid_train_and_learncurve_config_toml(test_configs_root):
    return test_configs_root.joinpath("invalid_train_and_learncurve_config.toml")


def test_train_and_learncurve_defined_raises(invalid_train_and_learncurve_config_toml):
    """test that a .toml config with both a TRAIN and a LEARNCURVE section raises a ValueError"""
    with pytest.raises(ValueError):
        vak.config.parse.from_toml_path(invalid_train_and_learncurve_config_toml)



"""tests for vak.config.spect_params module"""
import pytest

import vak.config.spect_params
import vak.split

from .attr_helpers import simple_attr


# approaching to testing validators adopted from
# https://github.com/python-attrs/attrs/blob/a025629e36440dcc27aee0ee5b04d6523bcc9931/tests/test_validators.py#L338
def test_freq_cutoffs_not_two_values_raises():
    """test that a ValueError is raised when freq_cutoffs is not just two values"""
    freq_cutoff_attrib = simple_attr("freq_cutoffs")

    with pytest.raises(ValueError):
        vak.config.spect_params.freq_cutoffs_validator(
            instance=None, attribute=freq_cutoff_attrib, value=[0]
        )

    with pytest.raises(ValueError):
        vak.config.spect_params.freq_cutoffs_validator(
            instance=None, attribute=freq_cutoff_attrib, value=[0, 10, 100]
        )


def test_invalid_transform_type_raises():
    """test that an invalid spectrogram transform type raises a ValueError"""
    transform_type_attrib = simple_attr("transform_type")

    with pytest.raises(ValueError):
        vak.config.spect_params.is_valid_transform_type(
            instance=None,
            attribute=transform_type_attrib,
            value="log",  # not in vak.config.spect_params.VALID_TRANSFORM_TYPES
        )


def test_freq_cutoffs_wrong_order_raises():
    """test that a ValueError is raised when freq_cutoffs[0] > freq_cutoffs[1]"""
    freq_cutoff_attrib = simple_attr("freq_cutoffs")

    with pytest.raises(ValueError):
        vak.config.spect_params.freq_cutoffs_validator(
            instance=None, attribute=freq_cutoff_attrib, value=[10000, 500]
        )


def test_spect_params_attrs_class(all_generated_configs_toml_path_pairs):
    """test that instantiating SpectParamsConfig class works as expected"""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if "SPECT_PARAMS" in config_toml:
            spect_params_section = config_toml["SPECT_PARAMS"]
            config = vak.config.spect_params.SpectParamsConfig(**spect_params_section)
            assert isinstance(config, vak.config.spect_params.SpectParamsConfig)



import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import functional as vakF

__all__ = ["dice_loss", "DiceLoss"]


# adapted from kornia (https://github.com/kornia/kornia/blob/master/kornia/losses/dice.py)
# originally based on https://github.com/kevinzakka/pytorch-goodies/blob/master/losses.py
def dice_loss(
    input: torch.Tensor, target: torch.Tensor, eps: float = 1e-8
) -> torch.Tensor:
    r"""Criterion that computes Sørensen-Dice Coefficient loss,
    adapted to work on a 1-dimensional time series.

    According to [1], we compute the Sørensen-Dice Coefficient as follows:
    .. math::
        \text{Dice}(x, class) = \frac{2 |X| \cap |Y|}{|X| + |Y|}
    Where:
       - :math:`X` scores that estimator assigns to each class.
       - :math:`Y` true class labels.

    the loss, is finally computed as:
    .. math::
        \text{loss}(x, class) = 1 - \text{Dice}(x, class)
    Reference:
        [1] https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
    Args:
        input (torch.Tensor): logits tensor with shape :math:`(N, C, T)` where C = number of classes,
          and T = number of timebins
        labels (torch.Tensor): labels tensor with shape :math:`(N, T)` where each value
          is :math:`0 ≤ targets[i] ≤ C−1`. Converted to one-hot vector with shape :math:`(N, C, T)`
          to compute the loss.
        eps (float, optional): Scalar to enforce numerical stabiliy. Default: 1e-8.
    Return:
        torch.Tensor: the computed loss.
    Example:
        >>> N = 5  # num_classes
        >>> input = torch.randn(1, N, 20, requires_grad=True)
        >>> target = torch.empty(1, 20, dtype=torch.long).random_(N)
        >>> output = dice_loss(input, target)
        >>> output.backward()
    """
    if not isinstance(input, torch.Tensor):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(type(input)))

    if not len(input.shape) == 3:
        raise ValueError(
            "Invalid input shape, should be 3 dimensions (N, C, T). "
            f"Got: {input.shape}"
        )

    if not input.shape[-1:] == target.shape[-1:]:
        raise ValueError(
            "Last dimension of input and target shapes must be the same size. "
            f"Got: {input.shape} and {target.shape}"
        )

    if not input.device == target.device:
        raise ValueError(
            f"input and target must be in the same device. Got: {input.device} and {target.device}"
        )

    # compute softmax over the classes axis
    input_soft: torch.Tensor = F.softmax(input, dim=1)

    # create the labels one hot tensor
    target_one_hot: torch.Tensor = vakF.one_hot(
        target, num_classes=input.shape[1], device=input.device, dtype=input.dtype
    )

    # compute the actual dice score
    dims = (1, 2)
    intersection = torch.sum(input_soft * target_one_hot, dims)
    cardinality = torch.sum(input_soft + target_one_hot, dims)

    dice_score = 2.0 * intersection / (cardinality + eps)

    return torch.mean(-dice_score + 1.0)


class DiceLoss(nn.Module):
    r"""Criterion that computes Sørensen-Dice Coefficient loss,
    adapted to work on a 1-dimensional time series.

    According to [1], we compute the Sørensen-Dice Coefficient as follows:
    .. math::
        \text{Dice}(x, class) = \frac{2 |X| \cap |Y|}{|X| + |Y|}
    Where:
       - :math:`X` scores that estimator assigns to each class.
       - :math:`Y` true class labels.
    the loss, is finally computed as:
    .. math::
        \text{loss}(x, class) = 1 - \text{Dice}(x, class)
    Reference:
        [1] https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
    Args:
        eps (float, optional): Scalar to enforce numerical stabiliy. Default: 1e-8.
    Shape:
        input (torch.Tensor): logits tensor with shape :math:`(N, C, T)` where C = number of classes,
          and T = number of timebins
        labels (torch.Tensor): labels tensor with shape :math:`(N, T)` where each value
          is :math:`0 ≤ targets[i] ≤ C−1`.
    Example:
        >>> N = 5  # num_classes
        >>> criterion = DiceLoss()
        >>> input = torch.randn(1, N, 3, 5, requires_grad=True)
        >>> target = torch.empty(1, 3, 5, dtype=torch.long).random_(N)
        >>> output = criterion(input, target)
        >>> output.backward()
    """

    def __init__(self, eps: float = 1e-8) -> None:
        super(DiceLoss, self).__init__()
        self.eps: float = eps

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        return dice_loss(input, target, self.eps)



"""module for functions that deal with vector of times from a spectrogram,
i.e. where elements are the times at bin centers"""
import numpy as np


def timebin_dur_from_vec(time_bins, n_decimals_trunc=5):
    """compute duration of a time bin, given the
    vector of time bin centers associated with a spectrogram

    Parameters
    ----------
    time_bins : numpy.ndarray
        vector of times in spectrogram, where each value is a bin center.
    n_decimals_trunc : int
        number of decimal places to keep when truncating the timebin duration calculated from
        the spectrogram arrays. Default is 5.

    Returns
    -------
    timebin_dur : float
        duration of a timebin, estimated from vector of times

    Notes
    -----
    takes mean of pairwise difference between neighboring time bins,
    to deal with floating point error, then rounds and truncates to specified decimal place
    """
    # first we round to the given number of decimals
    timebin_dur = np.around(np.mean(np.diff(time_bins)), decimals=n_decimals_trunc)
    # only after rounding do we truncate any decimal place past decade
    decade = 10 ** n_decimals_trunc
    timebin_dur = np.trunc(timebin_dur * decade) / decade
    return timebin_dur



from . import annotation


def to_map(labelset, map_unlabeled=True):
    """map set of labels to series of consecutive integers from 0 to n inclusive,
    where n is the number of labels in the set.

    This 'labelmap' is used when mapping labels from annotations of a vocalization into
    a label for every time bin in a spectrogram of that vocalization.

    If map_unlabeled is True, 'unlabeled' will be added to labelset, and will map to 0,
    so the total number of classes is n + 1.

    Parameters
    ----------
    labelset : set
        of labels used to annotate a Dataset.
    map_unlabeled : bool
        if True, include key 'unlabeled' in mapping. Any time bins in a spectrogram
        that do not have a label associated with them, e.g. a silent gap between vocalizations,
        will be assigned the integer that the 'unlabeled' key maps to.

    Returns
    -------
    labelmap : dict
        maps labels to integers
    """
    if type(labelset) != set:
        raise TypeError(f"type of labelset must be set, got type {type(labelset)}")

    labellist = []
    if map_unlabeled is True:
        labellist.append("unlabeled")

    labellist.extend(sorted(list(labelset)))

    labelmap = dict(zip(labellist, range(len(labellist))))
    return labelmap


def to_set(labels_list):
    """given a list of labels from annotations, return the set of (unique) labels

    Parameters
    ----------
    labels_list : list
         of lists, i.e. labels from annotations

    Returns
    -------
    labelset

    Examples
    --------
    >>> labels_list = [voc.annot.labels for voc in vds.voc_list]
    >>> labelset = to_set(labels_list)
    """
    all_labels = [lbl for labels in labels_list for lbl in labels]
    labelset = set(all_labels)
    return labelset


def from_df(vak_df):
    """returns labels for each vocalization in a dataset.
    Takes Pandas DataFrame representing the dataset, loads
    annotation for each row in the DataFrame, and then returns
    labels from each annotation.

    Parameters
    ----------
    vak_df : pandas.DataFrame
        created by vak.io.dataframe.from_files

    Returns
    -------
    labels : list
        of array-like, labels for each vocalization in the dataset.
    """
    annots = annotation.from_df(vak_df)
    return [annot.seq.labels for annot in annots]



from . import fixtures
from .fixtures import *


def pytest_addoption(parser):
    parser.addoption(
        "--models",
        action="store",
        default="teenytweetynet",
        nargs="+",
        help="vak models to test, space-separated list of names",
    )
    parser.addoption('--dtype', action="store", default="float32")


def pytest_generate_tests(metafunc):
    models = metafunc.config.option.models
    if isinstance(models, str):
        # wrap a single model name in a list
        models = [models]
    # **note!** fixture name is singular even though cmdopt is plural
    if "model" in metafunc.fixturenames and models is not None:
        metafunc.parametrize("model", models)

    dtype_names = None
    if 'dtype_name' in metafunc.fixturenames:
        raw_value = metafunc.config.getoption('--dtype')
        if raw_value == 'all':
            dtype_names = list(fixtures.torch.TEST_DTYPES.keys())
        else:
            dtype_names = raw_value.split(',')
        if dtype_names is not None:
            metafunc.parametrize('dtype_name', dtype_names)



"""functions for dealing with labeled timebin vectors"""
import numpy as np
import scipy.stats

from .timebins import timebin_dur_from_vec
from .validators import row_or_1d, column_or_1d


def has_unlabeled(labels_int, onsets_s, offsets_s, time_bins):
    """determine whether there are unlabeled segments in a spectrogram,
    given labels, onsets, and offsets of vocalizations, and vector of
    time bins from spectrogram

    Parameters
    ----------
    labels_int : list, numpy.ndarray
        a list or array of labels from the annotation for a vocalization,
        mapped to integers
    onsets_s : numpy.ndarray
        1d vector of floats, segment onsets in seconds
    offsets_s : numpy.ndarray
        1-d vector of floats, segment offsets in seconds
    time_bins : numpy.ndarray
        1-d vector of floats, time in seconds for center of each time bin of a spectrogram

    Returns
    -------
    has_unlabeled : bool
        if True, there are time bins that do not have labels associated with them
    """
    if (
        type(labels_int) == list
        and not all([type(lbl) == int for lbl in labels_int])
        or (
            type(labels_int) == np.ndarray
            and labels_int.dtype not in [np.int8, np.int16, np.int32, np.int64]
        )
    ):
        raise TypeError("labels_int must be a list or numpy.ndarray of integers")

    dummy_unlabeled_label = np.max(labels_int) + 1
    label_vec = np.ones((time_bins.shape[-1], 1), dtype="int8") * dummy_unlabeled_label
    onset_inds = [np.argmin(np.abs(time_bins - onset)) for onset in onsets_s]
    offset_inds = [np.argmin(np.abs(time_bins - offset)) for offset in offsets_s]
    for label, onset, offset in zip(labels_int, onset_inds, offset_inds):
        # offset_inds[ind]+1 because offset time bin is still "part of" syllable
        label_vec[onset : offset + 1] = label

    if dummy_unlabeled_label in label_vec:
        return True
    else:
        return False


def label_timebins(labels_int, onsets_s, offsets_s, time_bins, unlabeled_label=0):
    """makes a vector of labels for each time bin from a spectrogram,
    given labels, onsets, and offsets of vocalizations

    Parameters
    ----------
    labels_int : list, numpy.ndarray
        a list or array of labels from the annotation for a vocalization,
        mapped to integers
    onsets_s : numpy.ndarray
        1d vector of floats, segment onsets in seconds
    offsets_s : numpy.ndarray
        1-d vector of floats, segment offsets in seconds
    time_bins : mumpy.ndarray
        1-d vector of floats, time in seconds for center of each time bin of a spectrogram
    unlabeled_label : int
        label assigned to time bins that do not have labels associated with them.
        Default is 0

    Returns
    -------
    lbl_tb : numpy.ndarray
        same length as time_bins, with each element a label for each time bin
    """
    if (
        type(labels_int) == list
        and not all([type(lbl) == int for lbl in labels_int])
        or (
            type(labels_int) == np.ndarray
            and labels_int.dtype not in [np.int8, np.int16, np.int32, np.int64]
        )
    ):
        raise TypeError("labels_int must be a list or numpy.ndarray of integers")

    label_vec = np.ones((time_bins.shape[-1],), dtype="int8") * unlabeled_label
    onset_inds = [np.argmin(np.abs(time_bins - onset)) for onset in onsets_s]
    offset_inds = [np.argmin(np.abs(time_bins - offset)) for offset in offsets_s]
    for label, onset, offset in zip(labels_int, onset_inds, offset_inds):
        # offset_inds[ind]+1 because offset time bin is still "part of" syllable
        label_vec[onset : offset + 1] = label

    return label_vec


ALPHANUMERIC = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
DUMMY_SINGLE_CHAR_LABELS = [
    # some large range of characters not typically used as labels
    chr(x) for x in range(162, 400)
]
# start with alphanumeric since more human readable;
# mapping can be arbitrary as long as it's consistent
DUMMY_SINGLE_CHAR_LABELS = (
    *ALPHANUMERIC,
    *DUMMY_SINGLE_CHAR_LABELS
)


# added to fix https://github.com/NickleDave/vak/issues/373
def _multi_char_labels_to_single_char(labels_mapping):
    """returns a copy of a ``labels_mapping`` where any
    labels with multiple characters are converted to
    single characters

    this makes it possible to correctly compute metrics
    like Levenshtein edit distance
    """
    current_str_labels = sorted(
        # sort to be extra sure we get same order every time
        # (even though OrderedDict is now default in Python).
        # Same order forces mapping to single characters to be deterministic across function calls.
        labels_mapping.keys()
    )
    new_labels_mapping = {}
    for dummy_label_ind, label_str in enumerate(current_str_labels):
        label_int = labels_mapping[label_str]
        if len(label_str) > 1:
            # replace with dummy label
            new_label_str = DUMMY_SINGLE_CHAR_LABELS[dummy_label_ind]
            new_labels_mapping[new_label_str] = label_int
        else:
            new_labels_mapping[label_str] = label_int
    return new_labels_mapping


def lbl_tb2labels(labeled_timebins, labels_mapping, spect_ID_vector=None):
    """converts output of network from label for each frame
    to one label for each continuous segment.

    This function is used when evaluating a model,
    to convert outputs to dummy strings that make it possible
    to compute string-based metrics, such as edit distance.

    It should **not** be used to convert predictions
    to annotations, because it can modify the
    ``labels_mapping`` so that metrics are correctly
    computed.

    For mapping outputs to string label predictions,
    use ``vak.labeled_timebins.lbl_tb2segments``.

    Parameters
    ----------
    labeled_timebins : ndarray
        where each element is a label for a time bin.
        Such an array is the output of the network.
    labels_mapping : dict
        that maps str labels to consecutive integers.
        The mapping is inverted to convert back to str labels.
    spect_ID_vector : ndarray
        of same length as labeled_timebins, where each element
        is an ID # for the spectrogram from which labeled_timebins
        was taken.
        If provided, used to split the converted labels back to
        a list of label str, with one for each spectrogram.
        Default is None, in which case the return value is one long str.

    Returns
    -------
    labels : str or list
        labeled_timebins mapped back to label str.
        If spect_ID_vector was provided, then labels is split into a list of str,
        where each str corresponds to predicted labels for each predicted
        segment in each spectrogram as identified by spect_ID_vector.
    """
    labeled_timebins = row_or_1d(labeled_timebins)
    idx = np.diff(labeled_timebins, axis=0).astype(bool)
    idx = np.insert(idx, 0, True)

    labels = labeled_timebins[idx]

    # remove 'unlabeled' label
    if "unlabeled" in labels_mapping:
        labels = labels[labels != labels_mapping["unlabeled"]]

    # replace any multiple character labels in mapping
    # with dummy single-character labels
    # so that we do not affect Levenshtein distance computation
    # see https://github.com/NickleDave/vak/issues/373
    if any([len(label) > 1 for label in labels_mapping.keys()]):  # only re-map if necessary
        # (to minimize chance of knock-on bugs)
        labels_mapping = _multi_char_labels_to_single_char(labels_mapping)

    # only invert mapping and then map integer labels to characters
    # *after* ensuring all string labels are single-character
    inverse_labels_mapping = dict((v, k) for k, v in labels_mapping.items())
    labels = labels.tolist()
    labels = [inverse_labels_mapping[label] for label in labels]

    if spect_ID_vector:
        labels_list = []
        spect_ID_vector = spect_ID_vector[idx]
        labels_arr = np.asarray(labels)
        # need to split up labels by spect_ID_vector
        # this is probably not the most efficient way:
        spect_IDs = np.unique(spect_ID_vector)

        for spect_ID in spect_IDs:
            these = np.where(spect_ID_vector == spect_ID)
            curr_labels = labels_arr[these].tolist()
            if all([type(el) is str for el in curr_labels]):
                labels_list.append("".join(curr_labels))
            elif all([type(el) is int for el in curr_labels]):
                labels_list.append(curr_labels)
        return labels_list, spect_ID_vector
    else:
        if all([type(el) is str or type(el) is np.str_ for el in labels]):
            return "".join(labels)
        elif all([type(el) is int for el in labels]):
            return labels


def _segment_lbl_tb(lbl_tb):
    """helper function that segments vector of labeled timebins.

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector where each element represents a label for a timebin

    Returns
    -------
    labels : numpy.ndarray
        vector where each element is a label for a segment with its onset
        and offset indices given by the corresponding element in onset_inds
        and offset_inds.
    onset_inds : numpy.ndarray
        vector where each element is the onset index for a segment.
        Each onset corresponds to the value at the same index in labels.
    offset_inds : numpy.ndarray
        vector where each element is the offset index for a segment
        Each offset corresponds to the value at the same index in labels.
    """
    # factored out as a separate function to be able to test
    # and in case user wants to do just this with output of neural net
    offset_inds = np.where(np.diff(lbl_tb, axis=0))[0]
    onset_inds = offset_inds + 1
    offset_inds = np.concatenate((offset_inds, np.asarray([lbl_tb.shape[0] - 1])))
    onset_inds = np.concatenate((np.asarray([0]), onset_inds))
    labels = lbl_tb[onset_inds]
    return labels, onset_inds, offset_inds


def lbl_tb_segment_inds_list(lbl_tb, unlabeled_label=0):
    """given a vector of labeled timebins,
    returns a list of indexing vectors,
    one for each labeled segment in the vector.

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector of labeled timebins from spectrogram
    unlabeled_label : int
        label that was given to segments that were not labeled in annotation,
        e.g. silent periods between annotated segments. Default is 0.
    return_inds : bool
        if True, return list of indices for segments in lbl_tb, in addition to the segments themselves.
        if False, just return list of numpy.ndarrays that are the segments from lbl_tb.

    Returns
    -------
    segment_inds_list : list
        of numpy.ndarray, indices that will recover segments list from lbl_tb.
    """
    segment_inds = np.nonzero(lbl_tb != unlabeled_label)[0]
    return np.split(segment_inds, np.where(np.diff(segment_inds) != 1)[0] + 1)


def remove_short_segments(
    lbl_tb, segment_inds_list, timebin_dur, min_segment_dur, unlabeled_label=0
):
    """remove segments from vector of labeled timebins
    that are shorter than specified duration

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector of labeled spectrogram time bins, i.e.,
        where each element is a label for a time bin.
        Output of a neural network.
    segment_inds_list : list
        of numpy.ndarray, indices that will recover segments list from ``lbl_tb``.
        Returned by function ``vak.labels.lbl_tb_segment_inds_list``.
    timebin_dur : float
        Duration of a single timebin in the spectrogram, in seconds.
        Used to convert onset and offset indices in ``lbl_tb`` to seconds.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    unlabeled_label : int
        label that was given to segments that were not labeled in annotation,
        e.g. silent periods between annotated segments. Default is 0.

    Returns
    -------
    lbl_tb : numpy.ndarray
        with segments whose duration is shorter than ``min_segment_dur`` set to ``unlabeled_label``
    segment_inds_list : list
        of numpy.ndarray, with arrays removed that represented
        segments in ``lbl_tb`` that were shorter than ``min_segment_dur``
    """
    new_segment_inds_list = []

    for segment_inds in segment_inds_list:
        if segment_inds.shape[-1] * timebin_dur < min_segment_dur:
            lbl_tb[segment_inds] = unlabeled_label
            # DO NOT keep segment_inds array
        else:
            # do keep segment_inds array, don't change lbl_tb
            new_segment_inds_list.append(segment_inds)

    return lbl_tb, new_segment_inds_list


def majority_vote_transform(lbl_tb, segment_inds_list):
    """transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment.

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector of labeled spectrogram time bins, i.e.,
        where each element is a label for a time bin.
        Output of a neural network.
    segment_inds_list : list
        of numpy.ndarray, indices that will recover segments list from lbl_tb.
        Returned by funciton ``vak.labels.lbl_tb_segment_inds_list``.

    Returns
    -------
    lbl_tb : numpy.ndarray
        after the majority vote transform has been applied
    """
    for segment_inds in segment_inds_list:
        segment = lbl_tb[segment_inds]
        majority = scipy.stats.mode(segment)[0].item()
        lbl_tb[segment_inds] = majority

    return lbl_tb


def lbl_tb2segments(
    lbl_tb, labelmap, t, min_segment_dur=None, majority_vote=False, n_decimals_trunc=5
):
    """convert vector of labeled timebins into segments,
    by finding where continuous runs of a single label start
    and stop. Returns vectors of labels and onsets and offsets
    in units of seconds.

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector of labeled spectrogram time bins, i.e.,
        where each element is a label for a time bin.
        Output of a neural network.
    labelmap : dict
        that maps labels to consecutive integers.
        The mapping is inverted to convert back to labels.
    t : numpy.ndarray
        Vector of times; the times are bin centers of columns in a spectrogram.
        Returned by function that generated spectrogram.
        Used to convert onset and offset indices in lbl_tb to seconds.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        if True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.
    n_decimals_trunc : int
        number of decimal places to keep when truncating the timebin duration
        calculated from the vector of times t. Default is 5.

    Returns
    -------
    labels : numpy.ndarray
        vector where each element is a label for a segment with its onset
        and offset indices given by the corresponding element in onset_inds
        and offset_inds.
    onsets_s : numpy.ndarray
        vector where each element is the onset in seconds a segment.
        Each onset corresponds to the value at the same index in labels.
    offsets_s : numpy.ndarray
        vector where each element is the offset in seconds of a segment.
        Each offset corresponds to the value at the same index in labels.
    """
    lbl_tb = column_or_1d(lbl_tb)

    if "unlabeled" in labelmap:
        # handle the case when all time bins are predicted to be unlabeled
        # see https://github.com/NickleDave/vak/issues/383
        uniq_lbl_tb = np.unique(lbl_tb)
        if len(uniq_lbl_tb) == 1 and uniq_lbl_tb[0] == labelmap["unlabeled"]:
            return None, None, None

    timebin_dur = timebin_dur_from_vec(t, n_decimals_trunc)

    if min_segment_dur is not None or majority_vote:
        if "unlabeled" not in labelmap:
            raise ValueError(
                "min_segment_dur or majority_vote specified,"
                " but 'unlabeled' not in labelmap.\n"
                "Without 'unlabeled' segments these transforms cannot be applied."
            )
        segment_inds_list = lbl_tb_segment_inds_list(
            lbl_tb, unlabeled_label=labelmap["unlabeled"]
        )

    if min_segment_dur is not None:
        lbl_tb, segment_inds_list = remove_short_segments(
            lbl_tb,
            segment_inds_list,
            timebin_dur,
            min_segment_dur,
            labelmap["unlabeled"],
        )
        if len(segment_inds_list) == 0:  # no segments left after removing
            return None, None, None

    if majority_vote:
        lbl_tb = majority_vote_transform(lbl_tb, segment_inds_list)

    labels, onset_inds, offset_inds = _segment_lbl_tb(lbl_tb)

    # remove 'unlabeled' label
    if "unlabeled" in labelmap:
        keep = np.where(labels != labelmap["unlabeled"])[0]
        labels = labels[keep]
        onset_inds = onset_inds[keep]
        offset_inds = offset_inds[keep]

    # handle case where removing 'unlabeled' **after** clean-up leaves no segments
    if all([len(vec) == 0 for vec in (labels, onset_inds, offset_inds)]):
        return None, None, None

    inverse_labelmap = dict((v, k) for k, v in labelmap.items())
    labels = labels.tolist()
    labels = np.asarray([inverse_labelmap[label] for label in labels])
    # the 'best' estimate we can get of onset and offset times,
    # given binned times, and labels applied to each time bin,
    # is "some time" between the last labeled bin for one segment,
    # i.e. its offset, and the first labeled bin for the next
    # segment, i.e. its onset. In other words if the whole bin is labeled
    # as belonging to that segment, and the bin preceding it is labeled as
    # belonging to the previous section, then the onset of the current
    # segment must be the time between the two bins. To find those times
    # we use the bin centers and either subtract (for onsets) or add
    # (for offsets) half a timebin duration. This half a timebin
    # duration puts our onsets and offsets at the time "between" bins.
    onsets_s = t[onset_inds] - (timebin_dur / 2)
    offsets_s = t[offset_inds] + (timebin_dur / 2)

    # but this estimate will be "wrong" if we set the onset or offset time
    # outside the possible times in our timebin vector. Need to clean up.
    if onsets_s[0] < 0.0:
        onsets_s[0] = 0.0
    if offsets_s[-1] > t[-1]:
        offsets_s[-1] = t[-1]

    return labels, onsets_s, offsets_s



"""functions to plot learning curve results"""
from glob import glob
import os
import pickle
from configparser import ConfigParser

import joblib
import numpy as np
import matplotlib.pyplot as plt


def get_all_results_list(root, these_dirs_branch, config_file):
    these_dirs = [(root + this_dir[0], this_dir[1]) for this_dir in these_dirs_branch]
    config = ConfigParser()

    all_results_list = []

    for this_dir, bird_ID in these_dirs:
        results_dict = {}
        results_dict["dir"] = this_dir
        os.chdir(this_dir)
        config.read(config_file)
        results_dict["data_dir"] = config["DATA"]["data_dir"]
        results_dict["time_steps"] = config["NETWORK"]["time_steps"]
        results_dict["num_hidden"] = config["NETWORK"]["num_hidden"]
        results_dict["train_set_durs"] = [
            int(element) for element in config["TRAIN"]["train_set_durs"].split(",")
        ]
        with open(glob("summary*/train_err")[0], "rb") as f:
            results_dict["train_err"] = pickle.load(f)
        with open(glob("summary*/test_err")[0], "rb") as f:
            results_dict["test_err"] = pickle.load(f)
        pe = joblib.load(glob("summary*/y_preds_and_err_for_train_and_test")[0])
        results_dict["train_syl_err_rate"] = pe["train_syl_err_rate"]
        results_dict["test_syl_err_rate"] = pe["test_syl_err_rate"]
        results_dict["bird_ID"] = bird_ID
        all_results_list.append(results_dict)
        os.chdir("..")

    for el in all_results_list:
        el["mn_test_err"] = np.mean(el["test_err"], axis=1)
        el["mn_train_err"] = np.mean(el["train_err"], axis=1)
        el["mn_train_syl_err"] = np.mean(el["train_syl_err_rate"], axis=1)
        el["mn_test_syl_err"] = np.mean(el["test_syl_err_rate"], axis=1)
        el["train_set_durs"] = np.asarray(el["train_set_durs"])

    return all_results_list


def frame_error_rate(all_results_list):
    all_mn_test_err = []
    for el in all_results_list:
        all_mn_test_err.append(el["mn_test_err"])
    all_mn_test_err = np.asarray(all_mn_test_err)

    plt.style.use("ggplot")
    fig, ax = plt.subplots()
    for el in all_results_list:
        lbl = el["bird_ID"]
        ax.plot(
            el["train_set_durs"],
            el["mn_test_err"],
            label=lbl,
            linestyle="--",
            marker="o",
        )
    ax.plot(
        el["train_set_durs"],
        np.median(all_mn_test_err, axis=0),
        linestyle="--",
        marker="o",
        linewidth=3,
        color="k",
        label="median across birds",
    )

    fig.set_size_inches(16, 8)
    plt.legend(fontsize=20)
    plt.xticks(el["train_set_durs"])
    plt.tick_params(axis="both", which="major", labelsize=20, rotation=45)
    plt.title("Frame error rate as a function of training set size", fontsize=40)
    plt.ylabel("Frame error rate\nas measured on test set", fontsize=32)
    plt.xlabel("Training set size: duration in s", fontsize=32)
    plt.tight_layout()
    plt.savefig("frame-err-rate-v-train-set-size.png")


def syllable_error_rate(all_results_list):
    plt.style.use("ggplot")
    fig, ax = plt.subplots()
    fig.set_size_inches(16, 8)
    for el in all_results_list:
        lbl = el["bird_ID"]
        ax.plot(
            el["train_set_durs"],
            el["mn_test_syl_err"],
            label=lbl,
            linestyle=":",
            marker="o",
        )

    plt.scatter(120, 0.84, s=75)
    plt.text(
        75,
        0.7,
        "Koumura & Okanoya 2016,\n0.84 note error rate\nwith 120s training data",
        fontsize=20,
    )
    plt.scatter(480, 0.5, s=75)
    plt.text(
        355,
        0.35,
        "Koumura & Okanoya 2016,\n0.46 note error rate\nwith 480s training data",
        fontsize=20,
    )

    plt.legend(fontsize=20, loc="upper right")
    plt.title("Syllable error rate as a function of training set size", fontsize=40)
    plt.xticks(el["train_set_durs"])
    plt.tick_params(axis="both", which="major", labelsize=20, rotation=45)
    plt.ylabel("Syllable error rate\nas measured on test set", fontsize=32)
    plt.xlabel("Training set size: duration in s", fontsize=32)
    plt.tight_layout()
    plt.savefig("syl-error-rate-v-train-set-size.png")



"""Utilities for input validation

adapted in part from scikit-learn under license
https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/validation.py
"""
import warnings

import numpy as np


def column_or_1d(y, warn=False):
    """ravel column or 1d numpy array, else raise an error

    Parameters
    ----------
    y : array-like
    warn : boolean, default False
       To control display of warnings.

    Returns
    -------
    y : array
    """
    shape = np.shape(y)
    if len(shape) == 1:
        return np.ravel(y)
    if len(shape) == 2 and shape[1] == 1:
        if warn:
            warnings.warn(
                "A column-vector y was passed when a 1d array was"
                " expected. Please change the shape of y to "
                "(n_samples, ), for example using ravel().",
                stacklevel=2,
            )
        return np.ravel(y)

    raise ValueError("bad input shape {0}".format(shape))


def row_or_1d(y, warn=False):
    """ravel row or 1d numpy array, else raise an error

    Parameters
    ----------
    y : array-like
    warn : boolean, default False
       To control display of warnings.

    Returns
    -------
    y : array
    """
    shape = np.shape(y)
    if len(shape) == 1:
        return np.ravel(y)
    if len(shape) == 2 and shape[0] == 1:
        if warn:
            warnings.warn(
                "A row-vector y was passed when a 1d array was"
                " expected. Please change the shape of y to "
                "(n_samples, ), for example using ravel().",
                stacklevel=2,
            )
        return np.ravel(y)

    raise ValueError("bad input shape {0}".format(shape))



import pytest

import torch

import vak.nn.functional as F


# adapted from kornia, https://github.com/kornia/kornia/blob/master/test/utils/test_one_hot.py
def test_onehot():
    num_classes = 4
    labels = torch.zeros(2, 2, 1, dtype=torch.int64)
    labels[0, 0, 0] = 0
    labels[0, 1, 0] = 1
    labels[1, 0, 0] = 2
    labels[1, 1, 0] = 3

    # convert labels to one hot tensor
    one_hot = F.one_hot(labels, num_classes)

    assert 1.0 == pytest.approx(one_hot[0, labels[0, 0, 0], 0, 0].item())
    assert 1.0 == pytest.approx(one_hot[0, labels[0, 1, 0], 1, 0].item())
    assert 1.0 == pytest.approx(one_hot[1, labels[1, 0, 0], 0, 0].item())
    assert 1.0 == pytest.approx(one_hot[1, labels[1, 1, 0], 1, 0].item())



import pytest


@pytest.fixture
def events_path(generated_results_data_root):
    events_paths = sorted(
        generated_results_data_root.joinpath("train").glob("**/*events*")
    )
    assert len(events_paths) > 0
    return events_paths[0]



"""test loss functions"""
import torch
from torch.autograd import gradcheck
from torch.testing import assert_allclose

import vak.nn.loss


def tensor_to_gradcheck_var(tensor, dtype=torch.float64, requires_grad=True):
    """Converts the input tensor to a valid variable to check the gradient.
    `gradcheck` needs 64-bit floating point and requires gradient.
    """
    assert torch.is_tensor(tensor), type(tensor)
    return tensor.requires_grad_(requires_grad).type(dtype)


# adapted from kornia, https://github.com/kornia/kornia/blob/master/test/test_losses.py
class TestDiceLoss:
    def test_smoke(self, device, dtype):
        num_classes = 3
        logits = torch.rand(2, num_classes, 20, device=device, dtype=dtype)
        labels = torch.rand(2, 20) * num_classes
        labels = labels.to(device).long()

        criterion = vak.nn.loss.DiceLoss()
        assert criterion(logits, labels) is not None

    def test_all_zeros(self, device, dtype):
        num_classes = 3
        logits = torch.zeros(2, num_classes, 20, device=device, dtype=dtype)
        logits[:, 0] = 10.0
        logits[:, 1] = 1.0
        logits[:, 2] = 1.0
        labels = torch.zeros(2, 20, device=device, dtype=torch.int64)

        criterion = vak.nn.loss.DiceLoss()
        loss = criterion(logits, labels)
        assert_allclose(loss, torch.zeros_like(loss), rtol=1e-3, atol=1e-3)

    def test_gradcheck(self, device, dtype):
        num_classes = 3
        logits = torch.rand(2, num_classes, 20, device=device, dtype=dtype)
        labels = torch.rand(2, 20) * num_classes
        labels = labels.to(device).long()

        logits = tensor_to_gradcheck_var(logits)  # to var
        assert gradcheck(vak.nn.dice_loss, (logits, labels), raise_exception=True)

    def test_jit(self, device, dtype):
        num_classes = 3
        logits = torch.rand(2, num_classes, 20, device=device, dtype=dtype)
        labels = torch.rand(2, 20) * num_classes
        labels = labels.to(device).long()

        op = vak.nn.dice_loss
        op_script = torch.jit.script(op)

        assert_allclose(op(logits, labels), op_script(logits, labels))

    def test_module(self, device, dtype):
        num_classes = 3
        logits = torch.rand(2, num_classes, 20, device=device, dtype=dtype)
        labels = torch.rand(2, 20) * num_classes
        labels = labels.to(device).long()

        op = vak.nn.dice_loss
        op_module = vak.nn.loss.DiceLoss()

        assert_allclose(op(logits, labels), op_module(logits, labels))



import pytest

from vak.split.algorithms.validate import validate_split_durations


@pytest.mark.parametrize(
    "train_dur_in, val_dur_in, test_dur_in, dataset_dur, train_dur_expected, val_dur_expected, test_dur_expected",
    [
        (100, 25, 75, 200, 100, 25, 75),
        (100, None, -1, 200, 100, 0, -1),
        (-1, None, 100, 200, -1, 0, 100),
        (100, 20, -1, 200, 100, 20, -1),
        (-1, 20, 100, 200, -1, 20, 100),
        (100, None, None, 200, 100, 0, 0),
        (None, None, 100, 200, 0, 0, 100),
    ],
)
def test_validate_durs(
    train_dur_in,
    val_dur_in,
    test_dur_in,
    dataset_dur,
    train_dur_expected,
    val_dur_expected,
    test_dur_expected,
):
    train_dur_out, val_dur_out, test_dur_out = validate_split_durations(
        train_dur_in, val_dur_in, test_dur_in, dataset_dur
    )
    assert all(
        [
            train_dur_out == train_dur_expected,
            val_dur_out == val_dur_expected,
            test_dur_out == test_dur_expected,
        ]
    )


def test_validate_durs_all_durs_none_raises():
    train_dur_in = None
    val_dur_in = None
    test_dur_in = None
    dataset_dur = 200
    with pytest.raises(ValueError):
        # because we have to specify at least one of train_dur or test_dur
        validate_split_durations(train_dur_in, val_dur_in, test_dur_in, dataset_dur)


def test_validate_durs_val_only_raises():
    train_dur_in = None
    val_dur_in = 100
    test_dur_in = None
    dataset_dur = 200
    # because we only specified duration for validation set
    with pytest.raises(ValueError):
        validate_split_durations(train_dur_in, val_dur_in, test_dur_in, dataset_dur)


def test_validate_durs_negative_dur_raises():
    train_dur_in = -2
    test_dur_in = None
    val_dur_in = 100
    dataset_dur = 200
    # because negative duration is invalid
    with pytest.raises(ValueError):
        validate_split_durations(train_dur_in, val_dur_in, test_dur_in, dataset_dur)


def test_validate_durs_total_splits_greater_than_dataset_duration_raises():
    train_dur_in = 100
    test_dur_in = 100
    val_dur_in = 100
    dataset_dur = 200
    # because total splits duration is greater than dataset duration
    with pytest.raises(ValueError):
        validate_split_durations(train_dur_in, val_dur_in, test_dur_in, dataset_dur)



import pytest

import vak.converters


@pytest.mark.parametrize(
    ("toml_value", "labelset"),
    [
        ("abc", {"a", "b", "c"}),
        ("1235", {"1", "2", "3", "5"}),
        ("range: 1-3, 5", {"1", "2", "3", "5"}),
        ([1, 2, 3, 5], {"1", "2", "3", "5"}),
        (["a", "b", "c"], {"a", "b", "c"}),
        (["range: 1-3", "noise"], {"1", "2", "3", "noise"}),
    ],
)
def test_labelset_from_toml_value(toml_value, labelset):
    """test that function behaves as specified in docstring"""
    assert vak.converters.labelset_to_set(toml_value) == labelset


def test_labelset_from_toml_value_raises():
    """test that an invalid value in a labelset list raises a TypeError"""
    labelset_with_float = [1, "a", 0.3]
    with pytest.raises(TypeError):
        _ = vak.converters.labelset_to_set(labelset_with_float)



import json
from pathlib import Path
import shutil

import joblib
import pandas as pd
import torch.utils.data

from .. import csv
from .. import labels
from .. import models
from .. import tensorboard
from .. import transforms
from ..datasets.window_dataset import WindowDataset
from ..datasets.vocal_dataset import VocalDataset
from ..device import get_default as get_default_device
from ..io import dataframe
from ..logging import log_or_print
from ..paths import generate_results_dir_name_as_path


def train(
    model_config_map,
    csv_path,
    window_size,
    batch_size,
    num_epochs,
    num_workers,
    labelset=None,
    labelmap_path=None,
    checkpoint_path=None,
    spect_scaler_path=None,
    root_results_dir=None,
    results_path=None,
    spect_key="s",
    timebins_key="t",
    normalize_spectrograms=True,
    spect_id_vector=None,
    spect_inds_vector=None,
    x_inds=None,
    shuffle=True,
    val_step=None,
    ckpt_step=None,
    patience=None,
    device=None,
    logger=None,
):
    """Train models and save results.

    Saves checkpoint files for models,
    label map, and spectrogram scaler.
    These are saved either in ``results_path``
    if specified, or a new directory
    made inside ``root_results_dir``.

    Parameters
    ----------
    model_config_map : dict
        where each key-value pair is model name : dict of config parameters
    csv_path : str
        path to where dataset was saved as a csv.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shonw to neural networks
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    labelset : set
        of str or int, the set of labels that correspond to annotated segments
        that a network should learn to segment and classify. Note that if there
        are segments that are not annotated, e.g. silent gaps between songbird
        syllables, then `vak` will assign a dummy label to those segments
        -- you don't have to give them a label here.
    labelmap_path : str, pathlib.Path
        path to 'labelmap.json' file, that contains a mapping from a ``labelset``
        to consecutive integer labels, e.g., one generated by a previous run of
        ``vak.core.train``. Used when continuing training. Default is None.
    checkpoint_path : str, pathlib.Path
        path to a checkpoint file, 
        e.g., one generated by a previous run of ``vak.core.train``.
        If specified, this checkpoint will be loaded into model.
        Used when continuing training. 
        Default is None, in which case a new model is initialized.
    spect_scaler_path : str, pathlib.Path
        path to a ``SpectScaler`` used to normalize spectrograms,
        e.g., one generated by a previous run of ``vak.core.train``.
        Used when continuing training, for example on the same dataset.
        Default is None.
    root_results_dir : str, pathlib.Path
        Root directory in which a new directory will be created
        where results will be saved.
    results_path : str, pathlib.Path
        Directory where results will be saved.
        If specified, this parameter overrides ``root_results_dir``.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    device : str
        Device on which to work with model + data.
        Default is None. If None, then a device will be selected with vak.split.get_default.
        That function defaults to 'cuda' if torch.cuda.is_available is True.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    spect_id_vector : numpy.ndarray
        Parameter for WindowDataset. Represents the 'id' of any spectrogram,
        i.e., the index into spect_paths that will let us load it.
        Default is None.
    spect_inds_vector : numpy.ndarray
        Parameter for WindowDataset. Same length as spect_id_vector
        but values represent indices within each spectrogram.
        Default is None.
    x_inds : numpy.ndarray
        Parameter for WindowDataset.
        Indices of each window in the dataset. The value at x[0]
        represents the start index of the first window; using that
        value, we can index into spect_id_vector to get the path
        of the spectrogram file to load, and we can index into
        spect_inds_vector to index into the spectrogram itself
        and get the window.
        Default is None.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.
    """
    log_or_print(
        f"Loading dataset from .csv path: {csv_path}", logger=logger, level="info"
    )
    dataset_df = pd.read_csv(csv_path)
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    if val_step and not dataset_df["split"].str.contains("val").any():
        raise ValueError(
            f"val_step set to {val_step} but dataset does not contain a validation set; "
            f"please run `vak prep` with a config.toml file that specifies a duration for the validation set."
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    if results_path:
        results_path = Path(results_path).expanduser().resolve()
        if not results_path.is_dir():
            raise NotADirectoryError(
                f"results_path not recognized as a directory: {results_path}"
            )
    else:
        results_path = generate_results_dir_name_as_path(root_results_dir)
        results_path.mkdir()

    timebin_dur = dataframe.validate_and_get_timebin_dur(dataset_df)
    log_or_print(
        f"Size of timebin in spectrograms from dataset, in seconds: {timebin_dur}",
        logger=logger,
        level="info",
    )

    # ---------------- load training data  -----------------------------------------------------------------------------
    log_or_print(f"using training dataset from {csv_path}", logger=logger, level="info")
    # below, if we're going to train network to predict unlabeled segments, then
    # we need to include a class for those unlabeled segments in labelmap,
    # the mapping from labelset provided by user to a set of consecutive
    # integers that the network learns to predict
    train_dur = dataframe.split_dur(dataset_df, "train")
    log_or_print(
        f"Total duration of training split from dataset (in s): {train_dur}",
        logger=logger,
        level="info",
    )

    if labelset is None and labelmap_path is None:
        raise ValueError(
            'Must specify ``labelset`` or provide ``labelmap_path``, but both were None.'
        )
    if labelset:
        has_unlabeled = csv.has_unlabeled(csv_path, labelset, timebins_key)
        if has_unlabeled:
            map_unlabeled = True
        else:
            map_unlabeled = False
        labelmap = labels.to_map(labelset, map_unlabeled=map_unlabeled)
        log_or_print(
            f"number of classes in labelmap: {len(labelmap)}", logger=logger, level="info"
        )
        # save labelmap in case we need it later
        with open(results_path.joinpath("labelmap.json"), "w") as f:
            json.dump(labelmap, f)
    elif labelmap_path:
        log_or_print(
            f"loading labelmap from path: {labelmap_path}", logger=logger, level="info"
        )
        with labelmap_path.open("r") as f:
            labelmap = json.load(f)
        # copy to new results_path
        with open(results_path.joinpath("labelmap.json"), "w") as f:
            json.dump(labelmap, f)

    if spect_scaler_path is not None and normalize_spectrograms:
        log_or_print(
            f"loading spect scaler from path: {spect_scaler_path}",
            logger=logger,
            level="info")
        spect_standardizer = joblib.load(spect_scaler_path)
        shutil.copy(spect_scaler_path, results_path)
    # get transforms just before creating datasets with them
    elif normalize_spectrograms and spect_scaler_path is None:
        log_or_print(
            f"no spect_scaler_path provided, not loading",
            logger=logger,
            level="info",
        )
        log_or_print("will normalize spectrograms", logger=logger, level="info")
        spect_standardizer = transforms.StandardizeSpect.fit_df(
            dataset_df, spect_key=spect_key
        )
        joblib.dump(spect_standardizer, results_path.joinpath("StandardizeSpect"))
    elif spect_scaler_path is not None and not normalize_spectrograms:
        raise ValueError('spect_scaler_path provided but normalize_spectrograms was False, these options conflict')
    else: 
        #not normalize_spectrograms and spect_scaler_path is None:
        log_or_print(
            "normalize_spectrograms is False and no spect_scaler_path was provided, will not standardize spectrograms",
            logger=logger, level="info")
        spect_standardizer = None
    transform, target_transform = transforms.get_defaults("train", spect_standardizer)

    train_dataset = WindowDataset.from_csv(
        csv_path=csv_path,
        x_inds=x_inds,
        spect_id_vector=spect_id_vector,
        spect_inds_vector=spect_inds_vector,
        split="train",
        labelmap=labelmap,
        window_size=window_size,
        spect_key=spect_key,
        timebins_key=timebins_key,
        transform=transform,
        target_transform=target_transform,
    )
    log_or_print(
        f"Duration of WindowDataset used for training, in seconds: {train_dataset.duration()}",
        logger=logger,
        level="info",
    )
    train_data = torch.utils.data.DataLoader(
        dataset=train_dataset,
        shuffle=shuffle,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # ---------------- load validation set (if there is one) -----------------------------------------------------------
    if val_step:
        item_transform = transforms.get_defaults(
            "eval",
            spect_standardizer,
            window_size=window_size,
            return_padding_mask=True,
        )
        val_dataset = VocalDataset.from_csv(
            csv_path=csv_path,
            split="val",
            labelmap=labelmap,
            spect_key=spect_key,
            timebins_key=timebins_key,
            item_transform=item_transform,
        )
        val_data = torch.utils.data.DataLoader(
            dataset=val_dataset,
            shuffle=False,
            # batch size 1 because each spectrogram reshaped into a batch of windows
            batch_size=1,
            num_workers=num_workers,
        )
        val_dur = dataframe.split_dur(dataset_df, "val")
        log_or_print(
            f"Total duration of validation split from dataset (in s): {val_dur}",
            logger=logger,
            level="info",
        )

        log_or_print(
            f"will measure error on validation set every {val_step} steps of training",
            logger=logger,
            level="info",
        )
    else:
        val_data = None

    if device is None:
        device = get_default_device()

    models_map = models.from_model_config_map(
        model_config_map,
        num_classes=len(labelmap),
        input_shape=train_dataset.shape,
        logger=logger,
    )
    for model_name, model in models_map.items():
        if checkpoint_path is not None:
            log_or_print(
                f"loading checkpoint for {model_name} from path: {checkpoint_path}",
                logger=logger,
                level="info",
            )
            model.load(checkpoint_path, device=device)

        results_model_root = results_path.joinpath(model_name)
        results_model_root.mkdir()
        ckpt_root = results_model_root.joinpath("checkpoints")
        ckpt_root.mkdir()
        log_or_print(f"training {model_name}", logger=logger, level="info")
        writer = tensorboard.get_summary_writer(
            log_dir=results_model_root, filename_suffix=model_name
        )
        model.summary_writer = writer
        model.fit(
            train_data=train_data,
            num_epochs=num_epochs,
            ckpt_root=ckpt_root,
            val_data=val_data,
            val_step=val_step,
            ckpt_step=ckpt_step,
            patience=patience,
            device=device,
        )



from pathlib import Path
import shutil

from .. import config
from .. import core
from .. import logging
from ..paths import generate_results_dir_name_as_path
from ..timenow import get_timenow_as_str


def train(toml_path):
    """train models using training set specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None

    Trains models, saves results in new directory within root_results_dir specified
    in config.toml file, and adds path to that new directory to config.toml file.
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml_path(toml_path)

    if cfg.train is None:
        raise ValueError(
            f"train called with a config.toml file that does not have a TRAIN section: {toml_path}"
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    results_path = generate_results_dir_name_as_path(cfg.train.root_results_dir)
    results_path.mkdir(parents=True)
    # copy config file into results dir now that we've made the dir
    shutil.copy(toml_path, results_path)

    # ---- set up logging ----------------------------------------------------------------------------------------------
    logger = logging.get_logger(
        log_dst=results_path,
        caller="train",
        timestamp=get_timenow_as_str(),
        logger_name=__name__,
    )
    logger.info("Logging results to {}".format(results_path))

    model_config_map = config.models.map_from_path(toml_path, cfg.train.models)

    core.train(
        model_config_map=model_config_map,
        csv_path=cfg.train.csv_path,
        labelset=cfg.prep.labelset,
        window_size=cfg.dataloader.window_size,
        batch_size=cfg.train.batch_size,
        num_epochs=cfg.train.num_epochs,
        num_workers=cfg.train.num_workers,
        checkpoint_path=cfg.train.checkpoint_path,
        spect_scaler_path=cfg.train.spect_scaler_path,
        labelmap_path=cfg.train.labelmap_path,
        results_path=results_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.train.normalize_spectrograms,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
        patience=cfg.train.patience,
        device=cfg.train.device,
        logger=logger,
    )



from . import learncurve
from .annot import annotation, plot_labels, plot_segments
from .spect import spect, spect_annot

__all__ = [
    "learncurve",
    "plot_labels",
    "plot_segments",
    "spect",
    "spect_annot",
]



"""parses [PREP] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .validators import (
    is_a_directory,
    is_a_file,
    is_audio_format,
    is_annot_format,
    is_spect_format,
)
from ..converters import expanded_user_path, labelset_to_set


def duration_from_toml_value(value):
    """converter for dataset split durations.
    If value is -1, that value is returned -- specifies "use the remainder of the dataset".
    Other values are converted to float when possible."""
    if value == -1:
        return value
    else:
        return float(value)


def is_valid_duration(instance, attribute, value):
    """validator for dataset split durations"""
    if type(value) not in {int, float}:
        raise TypeError(
            f"invalid type for {attribute} of {instance}: {type(value)}. Type should be float or int."
        )

    if value == -1:  # specifies "use the remainder of the dataset"
        # so it is valid, but other negative values are not
        return

    if not value >= 0:
        raise ValueError(
            f"value specified for {attribute} of {instance} must be greater than or equal to zero, was {value}"
        )


@attr.s
class PrepConfig:
    """class to represent [PREP] section of config.toml file

    Attributes
    ----------
    data_dir : str
        path to directory with files from which to make dataset
    output_dir : str
        Path to location where data sets should be saved. Default is None,
        in which case data sets are saved in the current working directory.
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}.
    spect_format : str
        format of files containg spectrograms as 2-d matrices.
        One of {'mat', 'npy'}.
    spect_output_dir : str
        path to directory where array files containing spectrograms
        should be saved, when generated from audio files.
        Default is None, in which case the spectrogram files
        are saved in ``data_dir`` by ``vak.io.dataframe.from_files``.
    annot_format : str
        format of annotations. Any format that can be used with the
        crowsetta library is valid.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    labelset : set
        of str or int, the set of labels that correspond to annotated segments
        that a network should learn to segment and classify. Note that if there
        are segments that are not annotated, e.g. silent gaps between songbird
        syllables, then `vak` will assign a dummy label to those segments
        -- you don't have to give them a label here.
        Value for ``labelset`` is converted to a Python ``set``
        using ``vak.config.converters.labelset_from_toml_value``.
        See help for that function for details on how to specify labelset.
    train_dur : float
        total duration of training set, in seconds. When creating a learning curve,
        training subsets of shorter duration (specified by the 'train_set_durs' option
        in the LEARNCURVE section of a config.toml file) will be drawn from this set.
    val_dur : float
        total duration of validation set, in seconds.
    test_dur : float
        total duration of test set, in seconds.
    """

    data_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)
    output_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)

    audio_format = attr.ib(validator=validators.optional(is_audio_format), default=None)
    spect_format = attr.ib(validator=validators.optional(is_spect_format), default=None)
    spect_output_dir = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_directory),
        default=None,
    )
    annot_file = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )
    annot_format = attr.ib(validator=validators.optional(is_annot_format), default=None)

    labelset = attr.ib(
        converter=converters.optional(labelset_to_set),
        validator=validators.optional(instance_of(set)),
        default=None,
    )

    train_dur = attr.ib(
        converter=converters.optional(duration_from_toml_value),
        validator=validators.optional(is_valid_duration),
        default=None,
    )
    val_dur = attr.ib(
        converter=converters.optional(duration_from_toml_value),
        validator=validators.optional(is_valid_duration),
        default=None,
    )
    test_dur = attr.ib(
        converter=converters.optional(duration_from_toml_value),
        validator=validators.optional(is_valid_duration),
        default=None,
    )

    def __attrs_post_init__(self):
        if self.audio_format is not None and self.spect_format is not None:
            raise ValueError(f"cannot specify audio_format and spect_format")

        if self.audio_format is None and self.spect_format is None:
            raise ValueError(f"must specify either audio_format or spect_format")



"""tests for vak.core.learncurve module"""
import vak.config
import vak.constants
import vak.core.learncurve
import vak.paths


def learncurve_output_matches_expected(cfg, model_config_map, results_path):
    assert results_path.joinpath("learning_curve.csv").exists()

    for train_set_dur in cfg.learncurve.train_set_durs:
        train_set_dur_root = results_path.joinpath(f"train_dur_{train_set_dur}s")
        assert train_set_dur_root.exists()

        for replicate_num in range(1, cfg.learncurve.num_replicates + 1):
            replicate_path = train_set_dur_root.joinpath(f"replicate_{replicate_num}")
            assert replicate_path.exists()

            prep_csv = sorted(replicate_path.glob("*prep*csv"))
            assert len(prep_csv) == 1

            assert replicate_path.joinpath("labelmap.json").exists()

            assert replicate_path.joinpath("spect_id_vector.npy").exists()
            assert replicate_path.joinpath("spect_inds_vector.npy").exists()
            assert replicate_path.joinpath("x_inds.npy").exists()

            if cfg.learncurve.normalize_spectrograms:
                assert replicate_path.joinpath("StandardizeSpect").exists()

            for model_name in model_config_map.keys():
                eval_csv = sorted(replicate_path.glob(f"eval_{model_name}*csv"))
                assert len(eval_csv) == 1

                model_path = replicate_path.joinpath(model_name)
                assert model_path.exists()

                tensorboard_log = sorted(
                    model_path.glob(f"events.out.tfevents.*{model_name}")
                )
                assert len(tensorboard_log) == 1

                checkpoints_path = model_path.joinpath("checkpoints")
                assert checkpoints_path.exists()
                assert checkpoints_path.joinpath("checkpoint.pt").exists()
                if cfg.learncurve.val_step is not None:
                    assert checkpoints_path.joinpath(
                        "max-val-acc-checkpoint.pt"
                    ).exists()

    return True


def test_learncurve(specific_config, tmp_path, model, device):
    options_to_change = {"section": "LEARNCURVE", "option": "device", "value": device}

    toml_path = specific_config(
        config_type="learncurve",
        model=model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )

    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.learncurve.models)
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    vak.core.learning_curve(
        model_config_map,
        train_set_durs=cfg.learncurve.train_set_durs,
        num_replicates=cfg.learncurve.num_replicates,
        csv_path=cfg.learncurve.csv_path,
        labelset=cfg.prep.labelset,
        window_size=cfg.dataloader.window_size,
        batch_size=cfg.learncurve.batch_size,
        num_epochs=cfg.learncurve.num_epochs,
        num_workers=cfg.learncurve.num_workers,
        root_results_dir=None,
        results_path=results_path,
        previous_run_path=cfg.learncurve.previous_run_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.learncurve.normalize_spectrograms,
        shuffle=cfg.learncurve.shuffle,
        val_step=cfg.learncurve.val_step,
        ckpt_step=cfg.learncurve.ckpt_step,
        patience=cfg.learncurve.patience,
        device=cfg.learncurve.device,
        logger=None,
    )

    assert learncurve_output_matches_expected(cfg, model_config_map, results_path)


def test_learncurve_no_results_path(specific_config, tmp_path, model, device):
    root_results_dir = tmp_path.joinpath("test_learncurve_no_results_path")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "LEARNCURVE", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="learncurve",
        model=model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )

    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.learncurve.models)

    vak.core.learning_curve(
        model_config_map,
        train_set_durs=cfg.learncurve.train_set_durs,
        num_replicates=cfg.learncurve.num_replicates,
        csv_path=cfg.learncurve.csv_path,
        labelset=cfg.prep.labelset,
        window_size=cfg.dataloader.window_size,
        batch_size=cfg.learncurve.batch_size,
        num_epochs=cfg.learncurve.num_epochs,
        num_workers=cfg.learncurve.num_workers,
        root_results_dir=cfg.learncurve.root_results_dir,
        results_path=None,
        previous_run_path=cfg.learncurve.previous_run_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.learncurve.normalize_spectrograms,
        shuffle=cfg.learncurve.shuffle,
        val_step=cfg.learncurve.val_step,
        ckpt_step=cfg.learncurve.ckpt_step,
        patience=cfg.learncurve.patience,
        device=cfg.learncurve.device,
        logger=None,
    )

    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert learncurve_output_matches_expected(cfg, model_config_map, results_path)



"""tests for vak.config.prep module"""
import vak.config.prep


def test_parse_prep_config_returns_PrepConfig_instance(
    all_generated_configs_toml_path_pairs,
):
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        prep_section = config_toml["PREP"]
        config = vak.config.prep.PrepConfig(**prep_section)
        assert isinstance(config, vak.config.prep.PrepConfig)



"""fixtures relating to annotation files"""
import crowsetta
import pytest
import toml


@pytest.fixture
def annot_file_yarden(source_test_data_root):
    return source_test_data_root.joinpath(
        "spect_mat_annot_yarden", "llb3", "llb3_annot_subset.mat"
    )


@pytest.fixture
def annot_list_yarden(annot_file_yarden):
    scribe = crowsetta.Transcriber(format="yarden")
    annot_list = scribe.from_file(annot_file_yarden)
    return annot_list


@pytest.fixture
def labelset_yarden():
    """labelset as it would be loaded from a toml file

    don't return a set because we need to use this to test functions that convert it to a set
    """
    return [
        str(an_int)
        for an_int in [1, 2, 3, 6, 7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 18, 19]
    ]


@pytest.fixture
def annot_dir_notmat(source_test_data_root):
    return source_test_data_root.joinpath("audio_cbin_annot_notmat", "gy6or6", "032312")


@pytest.fixture
def annot_files_notmat(annot_dir_notmat):
    return sorted(annot_dir_notmat.glob("*.not.mat"))


@pytest.fixture
def annot_list_notmat(annot_files_notmat):
    scribe = crowsetta.Transcriber(format="notmat")
    annot_list = scribe.from_file(annot_files_notmat)
    return annot_list


@pytest.fixture
def labelset_notmat(generated_test_configs_root):
    """labelset as it would be loaded from a toml file

    don't return a set because we need to use this to test functions that convert it to a set"""
    a_train_notmat_config = sorted(
        generated_test_configs_root.glob("*train*notmat*toml")
    )[
        0
    ]  # get first config.toml from glob list
    # doesn't really matter which config, they all have labelset
    with a_train_notmat_config.open("r") as fp:
        a_train_notmat_toml = toml.load(fp)
    labelset = a_train_notmat_toml["PREP"]["labelset"]
    return labelset


@pytest.fixture
def annot_file_birdsongrec(source_test_data_root):
    return source_test_data_root.joinpath(
        "audio_wav_annot_birdsongrec", "Bird0", "Annotation.xml"
    )


@pytest.fixture
def annot_list_birdsongrec(annot_file_birdsongrec):
    scribe = crowsetta.Transcriber(format="birdsong-recognition-dataset")
    annot_list = scribe.from_file(annot_file_birdsongrec)
    return annot_list


@pytest.fixture
def specific_annot_list(annot_list_notmat, annot_list_yarden, annot_list_birdsongrec):
    """factory fixture, returns a function that
    returns a fixture containing a list of Annotation objects,
    given a specified annotation format

    so that unit tests can be parameterized with annotation format names
    """
    FORMAT_ANNOT_LIST_FIXTURE_MAP = {
        "notmat": annot_list_notmat,
        "yarden": annot_list_yarden,
        "birdsong-recognition-dataset": annot_list_birdsongrec,
    }

    def _annot_list_factory(format):
        return FORMAT_ANNOT_LIST_FIXTURE_MAP[format]

    return _annot_list_factory


@pytest.fixture
def specific_labelset(labelset_yarden, labelset_notmat):
    def _specific_labelset(annot_format):
        if annot_format == "yarden":
            return labelset_yarden
        elif annot_format == "notmat":
            return labelset_notmat
        else:
            raise ValueError(f"invalid annot_format: {annot_format}")

    return _specific_labelset



import numpy as np
import pytest

import vak.files.spect
import vak.labeled_timebins


def test_has_unlabeled():
    labels_1 = [1, 1, 1, 1, 2, 2, 3, 3, 3]
    onsets_s1 = np.asarray([0, 2, 4, 6, 8, 10, 12, 14, 16])
    offsets_s1 = np.asarray([1, 3, 5, 7, 9, 11, 13, 15, 17])
    time_bins = np.arange(0, 18, 0.001)
    has_ = vak.labeled_timebins.has_unlabeled(
        labels_1, onsets_s1, offsets_s1, time_bins
    )
    assert has_

    labels_1 = [1, 1, 1, 1, 2, 2, 3, 3, 3]
    onsets_s1 = np.asarray([0, 2, 4, 6, 8, 10, 12, 14, 16])
    offsets_s1 = np.asarray(
        [1.999, 3.999, 5.999, 7.999, 9.999, 11.999, 13.999, 15.999, 17.999]
    )
    time_bins = np.arange(0, 18, 0.001)
    has_ = vak.labeled_timebins.has_unlabeled(
        labels_1, onsets_s1, offsets_s1, time_bins
    )
    assert has_ is False


@pytest.mark.parametrize(
    "labeled_timebins, labels_mapping, spect_ID_vector, expected_labels",
    [
        (np.array([0, 0, 1, 1, 0, 0, 2, 2, 0, 0]), {'unlabeled': 0, 'a': 1, 'b': 2}, None, 'ab'),
        (np.array([0, 0, 1, 1, 0, 0, 2, 2, 0, 0]), {'unlabeled': 0, '1': 1, '2': 2}, None, '12'),
        (np.array([0, 0, 21, 21, 0, 0, 22, 22, 0, 0]), {'unlabeled': 0, '21': 21, '22': 22}, None, 'AB'),
        (np.array([0, 0, 11, 11, 0, 0, 12, 12, 0, 0]), {'unlabeled': 0, '11': 11, '12': 12}, None, 'AB'),
    ]
)
def test_lbl_tb2labels(labeled_timebins, labels_mapping, spect_ID_vector, expected_labels):
    labels = vak.labeled_timebins.lbl_tb2labels(labeled_timebins, labels_mapping, spect_ID_vector)
    assert labels == expected_labels


def test_segment_lbl_tb():
    lbl_tb = np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0])
    labels, onset_inds, offset_inds = vak.labeled_timebins._segment_lbl_tb(lbl_tb)
    assert np.array_equal(labels, np.asarray([0, 1, 0]))
    assert np.array_equal(onset_inds, np.asarray([0, 4, 8]))
    assert np.array_equal(offset_inds, np.asarray([3, 7, 11]))


@pytest.mark.parametrize(
    "lbl_tb, seg_inds_list_expected",
    [
        (np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]), [np.array([4, 5, 6, 7])]),
        # assert works when segment is at start of lbl_tb
        (np.asarray([1, 1, 1, 1, 0, 0, 0, 0]), [np.array([0, 1, 2, 3])]),
        # assert works with multiple segments
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1, 0, 0]),
            [np.array([3, 4, 5]), np.array([9, 10, 11])],
        ),
        # assert works when a segment is at end of lbl_tb
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1]),
            [np.array([3, 4, 5]), np.array([9, 10, 11])],
        ),
    ],
)
def test_lbl_tb_segment_inds_list(lbl_tb, seg_inds_list_expected):
    UNLABELED = 0

    seg_inds_list = vak.labeled_timebins.lbl_tb_segment_inds_list(
        lbl_tb=lbl_tb, unlabeled_label=UNLABELED
    )
    assert np.array_equal(seg_inds_list, seg_inds_list_expected)


def test_remove_short_segments():
    UNLABELED = 0

    # should do nothing when a labeled segment has all the same labels
    lbl_tb = np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0])
    segment_inds_list = vak.labeled_timebins.lbl_tb_segment_inds_list(
        lbl_tb, unlabeled_label=UNLABELED
    )
    TIMEBIN_DUR = 0.001
    MIN_SEGMENT_DUR = 0.002
    lbl_tb_tfm, segment_inds_list_out = vak.labeled_timebins.remove_short_segments(
        lbl_tb,
        segment_inds_list,
        timebin_dur=TIMEBIN_DUR,
        min_segment_dur=MIN_SEGMENT_DUR,
        unlabeled_label=UNLABELED,
    )

    lbl_tb_expected = np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0])
    assert np.array_equal(lbl_tb_tfm, lbl_tb_expected)


@pytest.mark.parametrize(
    "lbl_tb_in, lbl_tb_expected",
    [
        # should do nothing when a labeled segment has all the same labels
        (
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]),
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]),
        ),
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1, 0, 0]),
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0]),
        ),
        # test MajorityVote works when there is no 'unlabeled' segment at start of vector
        (np.asarray([1, 1, 2, 1, 0, 0, 0, 0]), np.asarray([1, 1, 1, 1, 0, 0, 0, 0])),
        # test MajorityVote works when there is no 'unlabeled' segment at end of vector
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1]),
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1]),
        ),
        # test that a tie results in lowest value class winning, default behavior of scipy.stats.mode
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 2, 2]),
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1]),
        ),
    ],
)
def test_majority_vote(lbl_tb_in, lbl_tb_expected):
    UNLABELED = 0

    segment_inds_list = vak.labeled_timebins.lbl_tb_segment_inds_list(
        lbl_tb_in, unlabeled_label=UNLABELED
    )
    lbl_tb_maj_vote = vak.labeled_timebins.majority_vote_transform(
        lbl_tb_in, segment_inds_list
    )
    assert np.array_equal(lbl_tb_maj_vote, lbl_tb_expected)


MAX_ABS_DIFF = 0.003  # milliseconds


def test_lbl_tb2segments_recovers_onsets_offsets_labels():
    onsets_s = np.asarray([1.0, 3.0, 5.0, 7.0])
    offsets_s = np.asarray([2.0, 4.0, 6.0, 8.0])
    labelset = set(list("abcd"))
    labelmap = vak.labels.to_map(labelset)

    labels = np.asarray(["a", "b", "c", "d"])
    timebin_dur = 0.001
    total_dur_s = 10
    timebins = (
        np.asarray(range(1, int(total_dur_s / timebin_dur) + 1)) * timebin_dur
    )  # [0.001, 0.002, ..., 10.0]
    lbl_tb = np.zeros(timebins.shape, dtype="int8")
    for onset, offset, lbl in zip(onsets_s, offsets_s, labels):
        on_ind = np.nonzero(timebins == onset)[0].item()
        off_ind = np.nonzero(timebins == offset)[0].item()
        lbl_tb[on_ind : off_ind + 1] = labelmap[lbl]

    labels_out, onsets_s_out, offsets_s_out = vak.labeled_timebins.lbl_tb2segments(
        lbl_tb, labelmap, timebins
    )

    assert np.array_equal(labels, labels_out)
    assert np.all(np.abs(onsets_s - onsets_s_out) < MAX_ABS_DIFF)
    assert np.all(np.abs(offsets_s - offsets_s_out) < MAX_ABS_DIFF)


# skip these for now because they cause tests to fail for reasons unrelated
# to what the test is testing
SPECT_FILES_TO_SKIP = [
    "llb3_0071_2018_04_23_17_38_30.wav.mat",  # has zero duration between syllable segments, onsets 54 and 55
    # I assume the same issue is coming up with these other two
    "llb3_0074_2018_04_23_17_41_08.wav.mat",
    "llb3_0016_2018_04_23_15_18_14.wav.mat",
]


def test_lbl_tb2segments_recovers_onsets_offsets_labels_from_real_data(
    specific_dataframe,
    labelset_yarden,
    model,
):
    """test that ``lbl_tb2segments`` recovers onsets and offsets from real data"""
    vak_df = specific_dataframe(
        config_type="train", model=model, spect_format="mat", annot_format="yarden"
    )
    labelmap = vak.labels.to_map(set(labelset_yarden))

    spect_paths = vak_df["spect_path"].values
    annot_list = vak.annotation.from_df(vak_df)
    spect_annot_map = vak.annotation.source_annot_map(spect_paths, annot_list)

    TIMEBINS_KEY = "t"

    for spect_path, annot in spect_annot_map.items():
        # in general not good to have conditionals in tests
        # but neglecting these weird edge case files for now
        if any(
            spect_path.endswith(spect_file_to_skip)
            for spect_file_to_skip in SPECT_FILES_TO_SKIP
        ):
            continue

        lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
        timebins = vak.files.spect.load(spect_path)[TIMEBINS_KEY]

        lbl_tb = vak.labeled_timebins.label_timebins(
            lbls_int,
            annot.seq.onsets_s,
            annot.seq.offsets_s,
            timebins,
            unlabeled_label=labelmap["unlabeled"],
        )

        labels, onsets_s_out, offsets_s_out = vak.labeled_timebins.lbl_tb2segments(
            lbl_tb, labelmap, timebins
        )
        assert np.all(np.char.equal(labels, annot.seq.labels))
        assert np.all(np.abs(annot.seq.onsets_s - onsets_s_out) < MAX_ABS_DIFF)
        assert np.all(np.abs(annot.seq.offsets_s - offsets_s_out) < MAX_ABS_DIFF)


def test_lbl_tb2segments_majority_vote():
    labelmap = {
        "unlabeled": 0,
        "a": 1,
        "b": 2,
    }
    lbl_tb = np.array([0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 2, 2, 1, 0, 0])
    timebins = np.arange(1, lbl_tb.shape[0] + 1) * 0.001
    labels_out, onsets_s_out, offsets_s_out = vak.labeled_timebins.lbl_tb2segments(
        lbl_tb, labelmap, timebins, majority_vote=True
    )
    assert np.all(np.char.equal(labels_out, np.array(["a", "b"])))


def test_lbl_tb2segments_all_unlabeled():
    """test that ``lbl_tb2segments`` returns all ``None``s when
    all elements in the input vector ``lbl_tb`` are the ``unlabeled`` class"""
    labelmap = {
        "unlabeled": 0,
        "a": 1,
        "b": 2,
    }
    N_TIMEBINS = 4000  # just want some number that's on the order of size of a typical Bengalese finch song
    lbl_tb = np.zeros(N_TIMEBINS).astype(int)
    timebins = np.arange(1, lbl_tb.shape[0] + 1) * 0.001
    labels_out, onsets_s_out, offsets_s_out = vak.labeled_timebins.lbl_tb2segments(
        lbl_tb, labelmap, timebins, majority_vote=True
    )
    assert all([out is None for out in [labels_out, onsets_s_out, offsets_s_out]])


@pytest.mark.parametrize(
    'y_pred, timebin_dur, min_segment_dur, labelmap',
    [
        (np.array([0, 0, 0, 0, 0, 0, 7, 7, 3,
                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                   7, 7, 7, 7, 7, 7, 7, 0, 0, 0, 0, 0, 0, ]),
         0.002,
         0.025,
         {"unlabeled": 0, "a": 3, "b": 7}),
    ]
)
def test_lbl_tb2segments_min_seg_dur_makes_all_unlabeled(y_pred,
                                                         timebin_dur,
                                                         min_segment_dur,
                                                         labelmap):
    """test that ``lbl_tb2segments`` returns all ``None``s when
    removing all segments less than the minimum segment duration
    causes all elements in the input vector ``lbl_tb``
    to become the ``unlabeled`` class"""
    # TODO: assert that applying 'minimum segment duration' post-processing does what we expect
    # i.e. converts all elements to 'unlabeled'
    timebins = np.arange(1, y_pred.shape[0] + 1) * timebin_dur
    labels_out, onsets_s_out, offsets_s_out = vak.labeled_timebins.lbl_tb2segments(
        y_pred, labelmap, timebins, min_segment_dur=min_segment_dur, majority_vote=True
    )
    assert all([out is None for out in [labels_out, onsets_s_out, offsets_s_out]])



import torch


def get_default():
    """get default device for torch.

    Returns
    -------
    device : str
        'cuda' if torch.cuda.is_available() is True,
        and returns 'cpu' otherwise.
    """
    if torch.cuda.is_available():
        return "cuda"
    else:
        return "cpu"
